'''Provides shared constants for Base Callables'''

from nawah.classes import Attr

FILE_ATTRS = {
    'user': Attr.ID(),
    'doc': Attr.ID(),
    'attr': Attr.STR(),
    'file': Attr.ANY(),
    'create_time': Attr.DATETIME(),
}



'''Provides 'check_permissions' Utility'''

import copy
import logging
from typing import TYPE_CHECKING, cast

from nawah.exceptions import NotPermittedException

if TYPE_CHECKING:
    from nawah.classes import Func, Module
    from nawah.types import NawahEnv

logger = logging.getLogger('nawah')


def check_permissions(*, func: 'Func', env: 'NawahEnv'):
    '''Matches Permission Sets against current session user privileges. Returns 'Func' object
    query_mod, doc_mod of Permission Set matched. If failed to match any, raises'''

    module = cast('Module', func.module)

    logger.debug(
        'Attempting to check permissions for \'%s\'.\'%s\' against \'%s\'',
        module.name,
        func.name,
        env['session']['user'],
    )

    # Loop over Permissions Set to find a match and return its query_mod, doc_mod
    for permission in func.permissions:
        # Check for allow-all *
        if permission.privilege == '*':
            return (
                copy.deepcopy(permission.query_mod),
                copy.deepcopy(permission.doc_mod),
            )
        # Set variables to use for checking permission
        privilege_module = module.name
        privilege = permission.privilege
        # If Permission Set privilege is dot-notated value, use it to set module, privilege
        if '.' in permission.privilege:
            privilege_module, privilege = permission.privilege.split('.')

        # Skip Permission Set if privilege_module not in user prvileges
        if privilege_module not in env['session']['user']['privileges']:
            continue

        if privilege not in env['session']['user']['privileges'][privilege_module]:
            continue

        return (copy.deepcopy(permission.query_mod), copy.deepcopy(permission.doc_mod))

    # If no match, raise NotPermittedException
    raise NotPermittedException(module_name=module.name, func_name=func.name)



'''Provides 'Default' dataclass'''

from dataclasses import dataclass
from typing import Callable, MutableMapping, MutableSequence, Union


@dataclass(kw_only=True)
class Default:
    '''Default dataclass serves role of defining Default Instruction, which instructs Nawah to set
    value for. Callable as value for 'value' should accept any number of following args only:
    'attr_name',  'attr_type',  'attr_val',  'doc' '''

    value: Union[
        None,
        str,
        int,
        float,
        bool,
        MutableMapping,
        MutableSequence,
        Callable,
    ]

    def __post_init__(self):
        pass



'''Provides 'Job' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Protocol

if TYPE_CHECKING:
    from croniter import croniter

    from nawah.types import NawahEnv


@dataclass(kw_only=True)
class Job:
    '''Job dataclass serves role of defining items of 'jobs' Config Attr which
    are callabled, called periodically per cron-based schedule'''

    job: 'JobCallable'
    schedule: str
    prevent_disable: bool
    # System Attributes
    _cron_schedule: Optional['croniter'] = None
    _next_time: Optional[str] = None
    _disabled: bool = False

    def __post_init__(self):
        self._cron_schedule = None
        self._next_time = None
        self._disabled = None

    @property
    def cron_schedule(self):
        '''Getter for '_cron_schedule' attr'''
        return self._cron_schedule

    @cron_schedule.setter
    def cron_schedule(self, val):
        '''Setter for '_cron_schedule' attr'''
        self._cron_schedule = val

    @property
    def next_time(self):
        '''Getter for '_next_time' attr'''
        return self._next_time

    @next_time.setter
    def next_time(self, val):
        '''Setter for '_next_time' attr'''
        self._next_time = val

    @property
    def disabled(self):
        '''Getter for '_disabled' attr'''
        return self._disabled

    @disabled.setter
    def disabled(self, val):
        '''Setter for '_disabled' attr'''
        self._disabled = val


class JobCallable(Protocol):
    '''Provides type-hint for 'job' callable of 'Job' '''

    # pylint: disable=too-few-public-methods
    def __call__(
        self,
        env: 'NawahEnv',
    ) -> None:
        ...



'''Provides exceptions used in Nawah'''

from ._attr import (InvalidAttrTypeArgException, InvalidAttrTypeException,
                    InvalidAttrTypeRefException, JSONPathNotFoundException)
from ._call import (InvalidCallEndpointException, InvalidFuncException,
                    InvalidModuleException, NotPermittedException)
from ._config import ConfigException
from ._data import UnknownDeleteStrategyException
from ._func import (FuncException, InvalidDocAttrException,
                    InvalidDocException, InvalidQueryAttrException,
                    InvalidQueryException, MissingDocAttrException,
                    MissingQueryAttrException)
from ._query import InvalidQueryArgException, UnknownQueryArgException
from ._validate import InvalidAttrException, MissingAttrException
from ._var import InvalidLocaleException, InvalidVarException

__all__ = [
    'InvalidAttrTypeArgException',
    'InvalidAttrTypeException',
    'InvalidAttrTypeRefException',
    'JSONPathNotFoundException',
    'InvalidCallEndpointException',
    'InvalidFuncException',
    'InvalidModuleException',
    'NotPermittedException',
    'ConfigException',
    'UnknownDeleteStrategyException',
    'FuncException',
    'InvalidDocAttrException',
    'InvalidDocException',
    'InvalidQueryAttrException',
    'InvalidQueryException',
    'MissingDocAttrException',
    'MissingQueryAttrException',
    'InvalidQueryArgException',
    'UnknownQueryArgException',
    'InvalidAttrException',
    'MissingAttrException',
    'InvalidLocaleException',
    'InvalidVarException',
]



import datetime
import json
import logging
import re
from typing import TYPE_CHECKING, MutableMapping, MutableSequence, cast

import aiohttp.web
from multidict import MultiDict
from passlib.hash import pbkdf2_sha512
from requests_toolbelt.multipart import decoder

from nawah import data as Data
from nawah.classes import app_encoder
from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import InvalidAttrException

from .._call import call
from .._config import _compile_anon_session, _compile_anon_user
from .._validate import validate_doc

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import NawahDoc, NawahEnv

logger = logging.getLogger('nawah')


async def _http_handler(request: aiohttp.web.Request):
    headers = MultiDict(
        [
            ('Server', 'Nawah'),
            ('Powered-By', 'Nawah, https://nawah.masaar.com'),
            ('Access-Control-Allow-Origin', '*'),
            ('Access-Control-Allow-Methods', 'GET,POST,OPTIONS'),
            (
                'Access-Control-Allow-Headers',
                'Content-Type,X-Auth-Bearer,X-Auth-Token,X-Auth-App',
            ),
            ('Access-Control-Expose-Headers', 'Content-Disposition'),
        ]
    )

    logger.debug('Received new %s request: %s', request.method, request.match_info)

    if request.method == 'OPTIONS':
        return aiohttp.web.Response(
            status=200,
            headers=headers,
            body=app_encoder.encode(
                {
                    'status': 200,
                    'msg': 'OPTIONS request is allowed',
                }
            ),
        )

    # Check for IP quota
    if str(request.remote) not in Config.sys.ip_quota:
        Config.sys.ip_quota[str(request.remote)] = {
            'counter': Config.quota_ip_min,
            'last_check': datetime.datetime.utcnow(),
        }
    else:
        if (
            datetime.datetime.utcnow()
            - Config.sys.ip_quota[str(request.remote)]['last_check']
        ).seconds > 259:
            Config.sys.ip_quota[str(request.remote)][
                'last_check'
            ] = datetime.datetime.utcnow()
            Config.sys.ip_quota[str(request.remote)]['counter'] = Config.quota_ip_min
        else:
            if Config.sys.ip_quota[str(request.remote)]['counter'] - 1 <= 0:
                logger.warning(
                    'Denying \'%s\' request from \'%s\' for hitting IP quota',
                    request.method,
                    request.remote,
                )
                headers['Content-Type'] = 'application/json; charset=utf-8'
                return aiohttp.web.Response(
                    status=429,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            'status': 429,
                            'msg': 'You have hit calls quota from this IP',
                            'args': {'code': 'CORE_REQ_IP_QUOTA_HIT'},
                        }
                    ),
                )

            Config.sys.ip_quota[str(request.remote)]['counter'] -= 1

    module_name = request.url.parts[1]
    func_name = request.url.parts[2]
    request_args = dict(request.match_info.items())

    # Extract Args Sets based on request.method
    args_sets = Config.modules[module_name].funcs[func_name].query_attrs
    args_sets = cast(MutableSequence[MutableMapping[str, 'Attr']], args_sets)

    # Attempt to validate query as doc
    for args_set in args_sets or []:
        if len(args_set.keys()) == len(args_set.keys()) and sum(
            1 for arg in args_set.keys() if arg in args_set.keys()
        ) == len(args_set.keys()):
            # Check presence and validate all attrs in doc args
            try:
                exception: Exception
                validate_doc(mode='create', doc=request_args, attrs=args_set)  # type: ignore
            except InvalidAttrException as e:
                exception = e
                headers['Content-Type'] = 'application/json; charset=utf-8'
                return aiohttp.web.Response(
                    status=400,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            'status': 400,
                            'msg': f'{str(e)} for \'{request.method}\' request on module '
                            '\'{module_name}\'',
                            'args': {'code': f'{module_name.upper()}_INVALID_ATTR'},
                        }
                    ).encode('utf-8'),
                )
            break

    conn = Data.create_conn()
    env: 'NawahEnv' = {
        'conn': conn,
        'REMOTE_ADDR': request.remote or 'localhost',
        'client_app': '__public',
        'args': {},
    }

    try:
        env['HTTP_USER_AGENT'] = request.headers['user-agent']
        env['HTTP_ORIGIN'] = request.headers['origin']
    except:
        env['HTTP_USER_AGENT'] = ''
        env['HTTP_ORIGIN'] = ''

    if 'X-Auth-Bearer' in request.headers or 'X-Auth-Token' in request.headers:
        logger.debug('Detected \'X-Auth\' header[s]')
        if (
            'X-Auth-Bearer' not in request.headers
            or 'X-Auth-Token' not in request.headers
            or 'X-Auth-App' not in request.headers
        ):
            logger.debug('Denying request due to missing \'X-Auth\' header')
            headers['Content-Type'] = 'application/json; charset=utf-8'
            return aiohttp.web.Response(
                status=400,
                headers=headers,
                body=app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'One \'X-Auth\' headers was set but not the other',
                    }
                ).encode('utf-8'),
            )
        # [TODO] Add condition to now allow a connection from a client with client_app = __sys
        if Config.client_apps and (
            request.headers['X-Auth-App'] not in Config.client_apps
            or (
                Config.client_apps[request.headers['X-Auth-App']].type == 'web'
                and env['HTTP_ORIGIN']
                not in Config.client_apps[request.headers['X-Auth-App']].origin
            )
        ):
            logger.debug('Denying request due to unauthorised client_app')
            headers['Content-Type'] = 'application/json; charset=utf-8'
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        'status': 403,
                        'msg': 'X-Auth headers could not be verified',
                        'args': {'code': 'CORE_SESSION_INVALID_XAUTH'},
                    }
                ).encode('utf-8'),
            )
        try:
            session_results = await call(
                'session/read',
                skip_events=[Event.PERM],
                env=env,
                query=[
                    {
                        '_id': request.headers['X-Auth-Bearer'],
                    }
                ],
            )
        except:
            headers['Content-Type'] = 'application/json; charset=utf-8'
            if Config.debug:
                return aiohttp.web.Response(
                    status=500,
                    headers=headers,
                    body=app_encoder.encode(
                        {
                            'status': 500,
                            'msg': f'Unexpected error has occurred [{str(exception)}]',
                            'args': {
                                'code': 'CORE_SERVER_ERROR',
                                'err': str(exception),
                            },
                        }
                    ).encode('utf-8'),
                )

            return aiohttp.web.Response(
                status=500,
                headers=headers,
                body=app_encoder.encode(
                    {
                        'status': 500,
                        'msg': 'Unexpected error has occurred',
                        'args': {'code': 'CORE_SERVER_ERROR'},
                    }
                ).encode('utf-8'),
            )

        if not session_results['args']['count'] or not pbkdf2_sha512.verify(
            request.headers['X-Auth-Token'],
            session_results['args']['docs'][0]['token_hash'],
        ):
            logger.debug('Denying request due to missing failed Call Authorisation')
            headers['Content-Type'] = 'application/json; charset=utf-8'
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(
                    {
                        'status': 403,
                        'msg': 'X-Auth headers could not be verified',
                        'args': {'code': 'CORE_SESSION_INVALID_XAUTH'},
                    }
                ).encode('utf-8'),
            )
        session = session_results['args']['docs'][0]
        session_results = await call(
            'session/reauth',
            skip_events=[Event.PERM],
            env=env,
            query=[
                {
                    '_id': request.headers['X-Auth-Bearer'],
                    'token': request.headers['X-Auth-Token'],
                    'groups': [],
                }
            ],
        )
        logger.debug('Denying request due to fail to reauth')
        if session_results['status'] != 200:
            headers['Content-Type'] = 'application/json; charset=utf-8'
            return aiohttp.web.Response(
                status=403,
                headers=headers,
                body=app_encoder.encode(session_results).encode('utf-8'),
            )

        session = session_results['args']['session']
    else:
        anon_user = _compile_anon_user()
        anon_session = _compile_anon_session()
        anon_session['user'] = anon_user
        session = anon_session

    env['session'] = session

    doc: 'NawahDoc' = {}

    if 'Content-Type' in request.headers:
        doc_content = await request.content.read()
        # Check Content-Type to decide how to deserialise content
        if request.headers['Content-Type'].startswith('multipart/form-data;'):
            multipart_content_type = request.headers['Content-Type']
            doc = {}
            for part in decoder.MultipartDecoder(
                doc_content, multipart_content_type
            ).parts:
                content_disposition = part.headers[b'Content-Disposition'].decode(
                    'UTF-8'
                )
                attr_name_match = re.findall('name="([^"]+)"', content_disposition)
                if not attr_name_match:
                    continue
                attr_name = attr_name_match[0]
                doc[attr_name] = part.content
                # For non-file values, decode unto str
                if 'filename' not in content_disposition:
                    doc[attr_name] = doc[attr_name].decode('UTF-8')
        elif request.headers['Content-Type'].startswith('application/json;'):
            doc = json.loads(doc_content)

    results = await call(
        f'{module_name}/{func_name}', env=env, query=[request_args], doc=doc
    )

    logger.debug('Closing connection')
    env['conn'].close()

    if 'return' not in results['args'] or results['args']['return'] == 'json':
        if 'return' in results['args']:
            del results['args']['return']
        headers['Content-Type'] = 'application/json; charset=utf-8'
        if results['status'] == 404:
            return aiohttp.web.Response(
                status=results['status'],
                headers=headers,
                body=app_encoder.encode(
                    {'status': 404, 'msg': 'Requested content not found'}
                ).encode('utf-8'),
            )
        return aiohttp.web.Response(
            status=results['status'],
            headers=headers,
            body=app_encoder.encode(results),
        )

    if results['args']['return'] == 'file':
        del results['args']['return']
        status = results['status']
        expiry_time = datetime.datetime.utcnow() + datetime.timedelta(days=30)
        headers['Last-Modified'] = str(results['args']['docs'][0]['lastModified'])
        headers['Content-Type'] = results['args']['docs'][0]['type']
        headers['Cache-Control'] = 'public, max-age=31536000'
        headers['Expires'] = expiry_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
        content = results['args']['docs'][0]['content']
        if 'If-Modified-Since' in request.headers:
            if request.headers['If-Modified-Since'] == headers['Last-Modified']:
                status = 304
                content = b''
        return aiohttp.web.Response(
            status=status,
            headers=headers,
            body=content,
        )

    if results['args']['return'] == 'msg':
        del results['args']['return']
        headers['Content-Type'] = 'application/json; charset=utf-8'
        return aiohttp.web.Response(
            status=results['status'], headers=headers, body=results['msg']
        )

    headers['Content-Type'] = 'application/json; charset=utf-8'
    return aiohttp.web.Response(
        status=405,
        headers=headers,
        body=app_encoder.encode({'status': 405, 'msg': 'METHOD NOT ALLOWED'}),
    )



'''Provides 'serve_app' Utility'''

import asyncio
import logging
import re
from dataclasses import fields
from typing import TYPE_CHECKING, Any, MutableMapping, MutableSequence, cast

from nawah.classes import App, Attr, Default, Var
from nawah.config import Config
from nawah.exceptions import ConfigException, InvalidVarException

from ._app import _run_app
from ._config import config_app, config_module
from ._val import deep_update, var_value
from ._validate import validate_attr, validate_type

if TYPE_CHECKING:
    from nawah.classes import Env, Package


logger = logging.getLogger('nawah')


def process_env_vars(config: 'Env', /):
    '''Checks all Var-prune Config Attrs of 'config' and updates value'''

    var_args: MutableMapping[str, MutableMapping[str, Any]] = {
        'debug': {'var': config.debug, 'type': bool, 'default': False},
        'port': {'var': config.port, 'type': int, 'default': 8081},
        'conn_timeout': {'var': config.conn_timeout, 'type': int},
        'quota_anon_min': {'var': config.quota_anon_min, 'type': int},
        'quota_auth_min': {'var': config.quota_auth_min, 'type': int},
        'quota_ip_min': {'var': config.quota_ip_min, 'type': int},
        'file_upload_limit': {
            'var': config.file_upload_limit,
            'type': int,
        },
        'file_upload_timeout': {
            'var': config.file_upload_timeout,
            'type': int,
        },
        'data_server': {'var': config.data_server, 'type': str},
        'data_name': {'var': config.data_name, 'type': str},
        'data_ssl': {'var': config.data_ssl, 'type': bool},
        'data_disk_use': {'var': config.data_disk_use, 'type': bool},
        'cache_server': {'var': config.cache_server, 'type': str},
        'cache_db': {
            'var': config.cache_db,
            'type': int,
            'default': 0,
        },
        'cache_username': {'var': config.cache_username, 'type': str},
        'cache_password': {'var': config.cache_password, 'type': str},
        'cache_expiry': {'var': config.cache_expiry, 'type': int},
        'error_reporting_server': {
            'var': config.error_reporting_server,
            'type': str,
        },
    }

    for arg_name, var_arg in var_args.items():
        if not isinstance(var_arg['var'], Var):
            continue
        try:
            setattr(
                config, arg_name, var_value(var_arg['var'], return_type=var_arg['type'])
            )
        except InvalidVarException as e:
            if 'default' not in var_arg:
                raise e
            setattr(config, arg_name, var_arg['default'])


def serve_app(app_config: 'App', /):
    '''Takes 'App' Object and setup runtime for it to run'''

    # pylint: disable=import-outside-toplevel

    from nawah.packages.core import core

    app_config_version = cast(str, app_config.version)

    if not re.match(r'^[0-9]+\.[0-9]+\.[0-9]+([ab][0-9]+)?$', app_config_version):
        raise ConfigException(
            f'App \'{app_config.name}\' defines invalid \'version\' \'{app_config.version}\''
        )

    # Check envs, env App Config Attr
    if app_config.envs and app_config.env:
        if isinstance(app_config.env, Var):
            logger.debug(
                'App \'{name}\' defines \'env\' App Config Attr as \'Var\' object. Getting its '
                'value..'
            )
            app_config.env = var_value(app_config.env)

        if app_config.env not in app_config.envs:
            raise ConfigException(
                'Value for \'env\' App Config Attr is \'{app_config.env}\', but'
                'no such Env is defined'
            )

        # Update runtime config with values from Env object
        app_config.env = cast(str, app_config.env)
        env = app_config.envs[app_config.env]
        for env_field in fields(env):
            if (env_attr_val := getattr(env, env_field.name)) is not None:
                # Specifically for l10n Config Attr, always update base value, and don't overwrite it
                if env_field.name == 'l10n':
                    # For a possibility app_config.l10n is None, set as L10N
                    if app_config.l10n is None:
                        app_config.l10n = {}
                    app_config.l10n.update(env_attr_val)
                else:
                    setattr(app_config, env_field.name, env_attr_val)

    if app_config.debug is True:
        logger.setLevel(logging.DEBUG)
        logger.debug('Set logging level to DEBUG (Config.debug==True)')

    if isinstance(app_config.debug, Var):
        try:
            if var_value(app_config.debug, return_type=lambda _: bool(int(_))):
                logger.setLevel(logging.DEBUG)
                logger.debug('Set logging level to DEBUG (Config.debug.var)')
                app_config.debug = True
        except InvalidVarException:
            logger.warning(
                'No value for \'Var\' object for \'debug\' App Config Attr. Logging '
                'remains at level ERROR'
            )

    logger.debug(
        'Getting value of \'Var\' object of \'admin_password\' App Config Attr..'
    )
    if isinstance(app_config.admin_password, Var):
        try:
            admin_password_val = var_value(app_config.admin_password)
            if not re.match(
                r'^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.{8,})', admin_password_val
            ):
                raise Exception(
                    f'Value \'{admin_password_val}\' of \'admin_password\' App Config Attr doesn\'t '
                    'match accepted password format'
                )
            app_config.admin_password = admin_password_val
            Config.sys.admin_check = True
        except InvalidVarException:
            logger.warning(
                'No value for \'Var\' object for \'admin_password\' App Config Attr. \'ADMIN\' doc '
                'will not be checked or updated'
            )
            Config.sys.admin_check = False

    if isinstance(app_config.anon_token, Var):
        logger.debug(
            'App \'{name}\' defines \'anon_token\' App Config Attr as \'Var\' object. Getting its '
            'value..'
        )
        app_config.anon_token = var_value(app_config.anon_token)

    process_env_vars(app_config)

    configs: MutableSequence['Package'] = [
        core,
        *(app_config.packages or []),
        app_config,
    ]

    for config in configs:
        for config_field in fields(config):
            attr_name = config_field.name
            attr_val = getattr(config, attr_name, None)

            if attr_val is None:
                continue

            if attr_name in ['envs', 'api_level', 'version', 'modules', 'packages']:
                continue

            # For vars_types Config Attr, preserve config name for debugging purposes
            if attr_name == 'vars_types':
                for var in attr_val:
                    config.name = cast(str, config.name)
                    Config.vars_types[var] = {
                        'package': config.name,
                        'type': attr_val[var],
                    }

            # Otherwise, update Config accroding to attr_val type
            elif isinstance(attr_val, list):
                for j in attr_val:
                    getattr(Config, attr_name).append(j)
                if attr_name == 'locales':
                    Config.locales = list(set(Config.locales))

            elif isinstance(attr_val, dict):
                if not getattr(Config, attr_name):
                    setattr(Config, attr_name, {})
                deep_update(target=getattr(Config, attr_name), new_values=attr_val)

            else:
                setattr(Config, attr_name, attr_val)

        if config.modules:
            config_modules = []
            for module in config.modules:
                if module.name in Config.modules:
                    raise ConfigException(
                        f'Module \'{module.name}\' exist in current runtime config'
                    )

                config_module(module_name=module.name, module=module)
                config_modules.append(module.name)
                Config.modules[module.name] = module

            Config.sys.packages[config.name or ''] = {
                'package': config,
                'modules': config_modules,
            }

        if getattr(config, 'packages', None):
            pass

        if isinstance(config, App):
            Config.sys.name = config.name or ''
            Config.sys.version = config.version or ''

    # [DOC] Update User, Session modules with populated attrs
    Config.modules['user'].attrs.update(Config.user_attrs)
    if sum(1 for attr in Config.user_settings if attr in Config.user_attrs) != 0:
        raise ConfigException(
            'At least one attr from \'user_settings\' is conflicting with an attr from '
            '\'user_attrs\''
        )
    for attr in Config.user_doc_settings:
        Config.modules['user'].attrs[attr] = Config.user_settings[attr].val_type
        Config.modules['user'].attrs[attr].default = Config.user_settings[attr].default
    Config.modules['user'].defaults['locale'] = Default(value=Config.locale)

    for user_attr_name, user_attr in Config.user_attrs.items():
        Config.modules['user'].unique_attrs.append(user_attr_name)
        Config.modules['user'].attrs[f'{user_attr_name}_hash'] = Attr.STR()
        session_auth_doc_args = cast(
            MutableSequence[MutableMapping[str, 'Attr']],
            Config.modules['session'].funcs['auth'].doc_attrs,
        )
        session_auth_doc_args.extend(
            [
                {
                    'hash': Attr.STR(),
                    user_attr_name: user_attr,
                    'groups': Attr.LIST(list=[Attr.ID()]),
                },
                {'hash': Attr.STR(), user_attr_name: user_attr},
            ]
        )

    # [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
    for var_name, var in Config.vars_types.items():
        if var_name not in Config.vars:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var, but not found in App '
                'Config'
            )
        try:
            validate_attr(
                mode='create',
                attr_name=var_name,
                attr_type=var['type'],
                attr_val=Config.vars[var_name],
            )
        except Exception as e:
            raise ConfigException(
                f'Package \'{var["package"]}\' requires \'{var_name}\' Var of type '
                f'\'{var["type"].type}\', but validation failed'
            ) from e

    # Validate Attr Type of type TYPE. Has to run before config_app so calls in config_app
    # depending on such attrs don't fail
    for attr_type in Config.sys.type_attrs:
        validate_type(attr_type=attr_type)

    asyncio.run(config_app())

    # Attempt to call post_config callable of all modules
    for module_name, module in Config.modules.items():
        logger.debug(
            'Attempting to check for \'post_config\' callable for module: %s',
            module_name,
        )
        if not module.post_config:
            logger.debug('- Module does not define \'post_config\' callable')
            continue

        logger.debug('- Module does define \'post_config\' callable. Calling it..')
        asyncio.run(module.post_config())

    _run_app()



'''Provides exceptions related to Nawah Function'''

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..types import ResultsArgs


class FuncException(Exception):
    '''Raised by \'call\' Utility when encounters an exception raised
    by Nawah Function callable'''

    def __init__(self, *, status: int, msg: str, args: 'ResultsArgs'):
        if 'code' not in args:
            raise Exception(
                'Invalid value for \'args\' for FuncException. Missing item \'code\''
            )

        super().__init__(msg, {'status': status, 'args': args})


class MissingQueryAttrException(Exception):
    '''Raied by '_check_query_attrs' if 'attr_name' is missing from checked Query object'''

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingQueryAttrException.format_msg(attr_name=attr_name),
            {'attr_name': attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        '''Formats exception message'''

        return f'Missing attr \'{attr_name}\' from Query'


class InvalidQueryAttrException(Exception):
    '''Raised by '_check_query_attrs' if 'attr' has invalid value'''

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidQueryAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {'attr_name': attr_name, 'attr_type': attr_type, 'val_type': val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        '''Formats exception message'''

        return (
            f'Invalid attr \'{attr_name}\' of type \'{val_type}\' with required type '
            f'\'{attr_type}\''
        )


class InvalidQueryException(Exception):
    '''Raised by '_check_query_attrs', when fail to match all 'query_attrs' sets'''

    status = 400

    def __init__(self, *, query_attrs_sets):
        super().__init__(
            InvalidQueryException.format_msg(query_attrs_sets=query_attrs_sets),
            {'query_attrs_sets': query_attrs_sets},
        )

    @staticmethod
    def format_msg(*, query_attrs_sets):
        '''Formats exception message'''

        return f'Invalid Query. Matched against: {query_attrs_sets}'


class MissingDocAttrException(Exception):
    '''Raied by '_check_doc_attrs' if 'attr_name' is missing from checked Doc object'''

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingDocAttrException.format_msg(attr_name=attr_name),
            {'attr_name': attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        '''Formats exception message'''

        return f'Missing attr \'{attr_name}\' from Doc'


class InvalidDocAttrException(Exception):
    '''Raised by '_check_doc_attrs' if 'attr' has invalid value'''

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidDocAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {'attr_name': attr_name, 'attr_type': attr_type, 'val_type': val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        '''Formats exception message'''

        return (
            f'Invalid attr \'{attr_name}\' of type \'{val_type}\' with required type '
            f'\'{attr_type}\''
        )


class InvalidDocException(Exception):
    '''Raised by '_check_doc_attrs', when fail to match all 'doc_attrs' sets'''

    status = 400

    def __init__(self, *, doc_attrs_sets):
        super().__init__(
            InvalidDocException.format_msg(doc_attrs_sets=doc_attrs_sets),
            {'doc_attrs_sets': doc_attrs_sets},
        )

    @staticmethod
    def format_msg(*, doc_attrs_sets):
        '''Formats exception message'''

        return f'Invalid Doc. Matched against: {doc_attrs_sets}'



'''Provides Nawah Utilities which provide functionality for internal and external behaviours'''

from ._attr_type import (decode_attr_type, encode_attr_type, extract_attr,
                         generate_attr_val)
from ._cache import update_cache
from ._call import call
from ._serve_app import serve_app
from ._val import (camel_to_upper, deep_update, expand_val, extract_val,
                   set_val, var_value)
from ._validate import (validate_attr, validate_doc, validate_type,
                        validate_user_setting)

__all__ = [
    'decode_attr_type',
    'encode_attr_type',
    'extract_attr',
    'generate_attr_val',
    'update_cache',
    'call',
    'serve_app',
    'camel_to_upper',
    'deep_update',
    'expand_val',
    'extract_val',
    'set_val',
    'var_value',
    'validate_attr',
    'validate_doc',
    'validate_type',
    'validate_user_setting',
]



'''Provides types used in Nawah'''

from ._types import (AnalyticsEvents, AppPackage, IPQuota, NawahDoc, NawahEnv,
                     NawahEvents, NawahQuery, NawahQuerySpecial,
                     NawahQuerySpecialGroup, NawahSession, Results,
                     ResultsArgs)

__all__ = [
    'AnalyticsEvents',
    'AppPackage',
    'IPQuota',
    'NawahDoc',
    'NawahEnv',
    'NawahEvents',
    'NawahQuery',
    'NawahQuerySpecial',
    'NawahQuerySpecialGroup',
    'NawahSession',
    'Results',
    'ResultsArgs',
]



'''Provides 'update_file' Base Function callable'''

from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import InvalidAttrException
from nawah.utils import call

from .exceptions import UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, Results


async def update_file(
    *,
    module_name: str,
    env: 'NawahEnv',
    query: 'Query',
    doc: 'NawahDoc',
) -> 'Results':
    '''Updates doc with attr of list of files in module with new file'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='update_file'
        )

    # [TODO] Allow use dot-notated attr path in attr query attr
    file_attr_name = query['attr'][0]
    if (
        file_attr_name not in module.attrs
        or module.attrs[file_attr_name].type != AttrType.LIST
        or not module.attrs[file_attr_name].args['list'][0].type != AttrType.FILE
    ):
        raise InvalidAttrException(
            attr_name=file_attr_name,
            attr_type=module.attrs[file_attr_name].type
            if file_attr_name in module.attrs
            else None,
            val_type=type(doc[file_attr_name]),
        )

    update_results = await call(
        'base/update',
        module_name=module_name,
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
        doc={query['attr'][0]: {'$append': doc['file']}},
    )

    return update_results



'''Provides 'user' Nawah Core Module'''

from nawah.classes import Attr, Default, Func, Module, Perm, Var

from ._callables import (_add_group, _create, _delete_group, _read,
                         _read_privileges)

user = Module(
    name='user',
    desc='\'user\' module provides data type and controller for users in Nawah eco-system. The '
    'permissions of the module methods are designed to be as secure for exposed calls, and as '
    'flexible for privileged-access',
    collection='user_docs',
    attrs={
        'name': Attr.LOCALE(desc='Name of the user as \'LOCALE\''),
        'locale': Attr.LOCALES(desc='Default locale of the user'),
        'create_time': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of the doc creation'
        ),
        'login_time': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of the last login'
        ),
        'groups': Attr.LIST(
            desc='List of \'_id\' for every group the user is member of',
            list=[Attr.ID(desc='\'_id\' of Group doc the user is member of')],
        ),
        'privileges': Attr.KV_DICT(
            desc='Privileges of the user. These privileges are always available to the user'
            'regardless of whether groups user is part of have them or not',
            key=Attr.STR(),
            val=Attr.LIST(list=[Attr.STR()]),
        ),
        'status': Attr.LITERAL(
            desc='Status of the user to determine whether user has access to the app or not',
            literal=['active', 'banned', 'deleted', 'disabled_password'],
        ),
    },
    defaults={
        'login_time': Default(value=None),
        'status': Default(value='active'),
        'groups': Default(value=[]),
        'privileges': Default(value={}),
    },
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='admin'),
                Perm(privilege='read', query_mod={'_id': Var.SESSION('user')}),
            ],
            call_args={'skip_user_settings': Attr.BOOL()},
            callable=_read,
        ),
        'create': Func(permissions=[Perm(privilege='admin')], callable=_create),
        'update': Func(
            permissions=[
                Perm(privilege='admin', doc_mod={'groups': None}),
                Perm(
                    privilege='update',
                    query_mod={'_id': Var.SESSION('user')},
                    doc_mod={'groups': None, 'privileges': None},
                ),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='admin'),
                Perm(privilege='delete', query_mod={'_id': Var.SESSION('user')}),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
        'read_privileges': Func(
            permissions=[
                Perm(privilege='admin'),
                Perm(privilege='read', query_mod={'_id': Var.SESSION('user')}),
            ],
            query_attrs={'_id': Attr.ID()},
            callable=_read_privileges,
        ),
        'add_group': Func(
            permissions=[Perm(privilege='admin')],
            query_attrs={'_id': Attr.ID()},
            doc_attrs=[{'group': Attr.ID()}, {'group': Attr.LIST(list=[Attr.ID()])}],
            callable=_add_group,
        ),
        'delete_group': Func(
            permissions=[Perm(privilege='admin')],
            query_attrs={'_id': Attr.ID(), 'group': Attr.ID()},
            callable=_delete_group,
        ),
        'retrieve_file': Func(permissions=[Perm(privilege='__sys')], get_func=True),
        'create_file': Func(permissions=[Perm(privilege='__sys')]),
        'update_file': Func(permissions=[Perm(privilege='__sys')]),
        'delete_file': Func(permissions=[Perm(privilege='__sys')]),
    },
)



'''Provides 'validate_user_setting' Utility'''

from typing import TYPE_CHECKING

from nawah.classes import Attr

from ._type import validate_type

if TYPE_CHECKING:
    from nawah.classes import UserSetting

# [TODO] Use custom exceptions


def validate_user_setting(*, user_setting: 'UserSetting'):
    '''Validates attributes values of 'UserSettingDict' dict'''

    # Validate type
    if user_setting.type not in ['user', 'user_sys']:
        raise Exception(
            f'Invalid \'type\' of type \'{type(user_setting.type)}\' with required \'user\', or'
            ' \'user_sys\''
        )

    # Validate val_type
    if not isinstance(user_setting.val_type, Attr):
        raise Exception(
            f'Invalid \'val_type\' of type \'{type(user_setting.val_type)}\' with required type'
            ' \'Attr\''
        )

    # Validate val_type Attr Type
    validate_type(attr_type=user_setting.val_type)



import asyncio
import datetime
import json
import logging
import random
import traceback
from functools import partial
from typing import TYPE_CHECKING

import aiohttp.web

from nawah import data as Data
from nawah.classes import app_encoder
from nawah.config import Config

from .._call import call
from .._config import _compile_anon_session, _compile_anon_user
from .._val import camel_to_upper
from ._utils import _close_session

if TYPE_CHECKING:
    from asyncio.futures import Future

    from nawah.types import NawahEnv

logger = logging.getLogger('nawah')


async def _websocket_handler(request: aiohttp.web.Request):
    conn = Data.create_conn()
    logger.debug('Websocket connection starting with client at \'%s\'', request.remote)
    ws = aiohttp.web.WebSocketResponse()
    await ws.prepare(request)

    session_lock_key = random.random()

    while Config.sys.session_lock != session_lock_key:
        if Config.sys.session_lock is None:
            Config.sys.session_lock = session_lock_key

    Config.sys.session_counter += 1

    env: 'NawahEnv' = {
        'id': str(Config.sys.session_counter),
        'conn': conn,
        'REMOTE_ADDR': request.remote or 'localhost',
        'ws': ws,
        'init': False,
        'last_call': datetime.datetime.utcnow(),
        'quota': {
            'counter': Config.quota_anon_min,
            'last_check': datetime.datetime.utcnow(),
        },
        'args': {},
    }
    Config.sys.sessions[str(Config.sys.session_counter)] = env
    Config.sys.session_lock = None

    try:
        env['HTTP_USER_AGENT'] = request.headers['user-agent']
        env['HTTP_ORIGIN'] = request.headers['origin']
    except:
        env['HTTP_USER_AGENT'] = ''
        env['HTTP_ORIGIN'] = ''

    logger.debug(
        'Websocket connection #\'%s\' ready with client at \'%s\'',
        env['id'],
        env['REMOTE_ADDR'],
    )

    await ws.send_str(
        app_encoder.encode(
            {
                'status': 200,
                'msg': 'Connection ready',
                'args': {'code': 'CORE_CONN_READY'},
            }
        )
    )

    async for msg in ws:
        if 'conn' not in env:
            await ws.close()
            break
        logger.debug(
            'Received new message from session #\'%s\': %s', env['id'], msg.data[:256]
        )
        if msg.type == aiohttp.WSMsgType.TEXT:
            logger.debug(
                'Config.sys.ip_quota on session #\'%s\': %s',
                env['id'],
                Config.sys.ip_quota,
            )
            logger.debug(
                'session_quota: on session #\'%s\': %s', env['id'], env['quota']
            )
            # Check for IP quota
            if str(request.remote) not in Config.sys.ip_quota:
                Config.sys.ip_quota[str(request.remote)] = {
                    'counter': Config.quota_ip_min,
                    'last_check': datetime.datetime.utcnow(),
                }
            else:
                if (
                    datetime.datetime.utcnow()
                    - Config.sys.ip_quota[str(request.remote)]['last_check']
                ).seconds > 59:
                    Config.sys.ip_quota[str(request.remote)][
                        'last_check'
                    ] = datetime.datetime.utcnow()
                    Config.sys.ip_quota[str(request.remote)][
                        'counter'
                    ] = Config.quota_ip_min
                else:
                    # If hit quota, deny call
                    if Config.sys.ip_quota[str(request.remote)]['counter'] - 1 <= 0:
                        logger.warning(
                            'Denying Websocket request from \'%s\' for hitting IP quota',
                            request.remote,
                        )
                        asyncio.create_task(
                            _handle_msg(
                                env=env,
                                msg=msg,
                                decline_quota='ip',
                            )
                        )
                        continue
                    # Else, update quota counter
                    Config.sys.ip_quota[str(request.remote)]['counter'] -= 1
            # Check for session quota
            if (datetime.datetime.utcnow() - env['quota']['last_check']).seconds > 59:
                env['quota']['last_check'] = datetime.datetime.utcnow()
                env['quota']['counter'] = (
                    (Config.quota_anon_min - 1)
                    if not env['session']
                    or env['session']['token'] == Config.anon_token
                    else (Config.quota_auth_min - 1)
                )
                asyncio.create_task(_handle_msg(env=env, msg=msg))
            else:
                # If hit quota, deny call
                if env['quota']['counter'] - 1 <= 0:
                    asyncio.create_task(
                        _handle_msg(
                            env=env,
                            msg=msg,
                            decline_quota='session',
                        )
                    )
                    continue
                # Else, update quota counter
                env['quota']['counter'] -= 1
                asyncio.create_task(_handle_msg(env=env, msg=msg))

    if 'id' in env:
        await _close_session(env['id'])

    return ws


async def _handle_msg(
    env: 'NawahEnv',
    msg: aiohttp.WSMessage,
    decline_quota: str = None,
):
    try:
        env['last_call'] = datetime.datetime.utcnow()
        try:
            env['session']['token']
        except Exception:
            anon_user = _compile_anon_user()
            anon_session = _compile_anon_session()
            anon_session['user'] = anon_user
            env['session'] = anon_session
        res = json.loads(msg.data)

        # Check if msg should be denied for quota hit
        if decline_quota == 'ip':
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 429,
                        'msg': 'You have hit calls quota from this IP',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_REQ_IP_QUOTA_HIT',
                        },
                    }
                )
            )
            return

        if decline_quota == 'session':
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 429,
                        'msg': 'You have hit calls quota',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_REQ_SESSION_QUOTA_HIT',
                        },
                    }
                )
            )
            return

        logger.debug('Decoded request: %s', app_encoder.encode(res))

        if 'endpoint' not in res.keys():
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Request missing endpoint',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_REQ_NO_ENDPOINT',
                        },
                    }
                )
            )
            return

        if env['init'] is False:
            if res['endpoint'] != 'conn/verify':
                await env['ws'].send_str(
                    app_encoder.encode(
                        {
                            'status': 1008,
                            'msg': 'Request token is not accepted',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_REQ_NO_VERIFY',
                            },
                        }
                    )
                )
                await env['ws'].close()
                return

            # [TODO] Add condition to now allow a connection from a client with client_app = __sys
            if Config.client_apps and (
                'doc' not in res
                or 'app' not in res['doc']
                or res['doc']['app'] not in Config.client_apps
                or (
                    Config.client_apps[res['doc']['app']].type == 'web'
                    and env['HTTP_ORIGIN']
                    not in Config.client_apps[res['doc']['app']].origin
                )
            ):
                await env['ws'].send_str(
                    app_encoder.encode(
                        {
                            'status': 1008,
                            'msg': 'Request token is not accepted',
                            'args': {
                                'call_id': res['call_id']
                                if 'call_id' in res.keys()
                                else None,
                                'code': 'CORE_REQ_NO_VERIFY',
                            },
                        }
                    )
                )
                await env['ws'].close()
                return

            env['init'] = True
            # [TODO] Add condition to now allow a connection from a client with client_app = __sys
            if not Config.client_apps:
                env['client_app'] = '__public'
            else:
                env['client_app'] = res['doc']['app']
            logger.debug('Connection on session #\'%s\' is verified', env['id'])
            # if Config.analytics_events['app_conn_verified']:
            #     asyncio.create_task(
            #         call(
            #             'analytic/create',
            #             skip_events=[Event.PERM],
            #             env=env,
            #             doc={
            #                 'event': 'CONN_VERIFIED',
            #                 'subevent': env['client_app'],
            #                 'args': {
            #                     'REMOTE_ADDR': env['REMOTE_ADDR'],
            #                     'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
            #                 },
            #             },
            #         )
            #     )
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 200,
                        'msg': 'Connection established',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_CONN_OK',
                        },
                    }
                )
            )
            return

        if res['endpoint'] == 'conn/close':
            logger.debug(
                f'Received connection close instructions on session #\'{env["id"]}\''
            )
            await env['ws'].close()
            return

        if res['endpoint'] == 'heart/beat':
            logger.debug(f'Received connection heartbeat on session #\'{env["id"]}\'')
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 200,
                        'msg': 'Heartbeat received',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_HEARTBEAT_OK',
                        },
                    }
                )
            )
            return

        res['endpoint'] = res['endpoint']

        if (
            res['endpoint'] in ['session/auth', 'session/reauth']
            and str(env['session']['_id']) != 'f00000000000000000000012'
        ):
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'You are already authed',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_SESSION_ALREADY_AUTHED',
                        },
                    }
                )
            )
            return

        if (
            res['endpoint'] == 'session/signout'
            and str(env['session']['_id']) == 'f00000000000000000000012'
        ):
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Singout is not allowed for \'__ANON\' user',
                        'args': {
                            'call_id': res['call_id']
                            if 'call_id' in res.keys()
                            else None,
                            'code': 'CORE_SESSION_ANON_SIGNOUT',
                        },
                    }
                )
            )
            return

        if 'query' not in res.keys():
            res['query'] = []
        if 'doc' not in res.keys():
            res['doc'] = {}
        if 'call_id' not in res.keys():
            res['call_id'] = ''

        request = {
            'call_id': res['call_id'],
            # 'sid': res['sid'] or False,
            'query': res['query'],
            'doc': res['doc'],
            'path': res['endpoint'].split('/'),
        }

        if len(request['path']) != 2:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Endpoint path is invalid',
                        'args': {
                            'call_id': request['call_id'],
                            'code': 'CORE_REQ_INVALID_PATH',
                        },
                    }
                )
            )
            return

        module = request['path'][0]

        if module not in Config.modules:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Endpoint module is invalid',
                        'args': {
                            'call_id': request['call_id'],
                            'code': 'CORE_REQ_INVALID_MODULE',
                        },
                    }
                )
            )
            return

        if request['path'][1] not in Config.modules[module].funcs:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Endpoint is invalid',
                        'args': {
                            'call_id': request['call_id'],
                            'code': 'CORE_REQ_INVALID_METHOD',
                        },
                    }
                )
            )
            return

        if Config.modules[module].funcs[request['path'][1]].get_func:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 400,
                        'msg': 'Endpoint is a GET Function',
                        'args': {
                            'call_id': request['call_id'],
                            'code': 'CORE_REQ_GET_FUNC',
                        },
                    }
                )
            )
            return

        # if not request['sid']:
        request['sid'] = 'f00000000000000000000012'

        query = request['query']
        doc = request['doc']
        # await _process_file_obj(doc=doc, env=env)
        call_task = asyncio.create_task(
            call(
                f'{module}/{request["path"][1]}',
                skip_events=[],
                env=env,
                query=query,
                doc=doc,
            )
        )
        call_task.add_done_callback(
            partial(_call_callback, env=env, call_id=request['call_id'])
        )

    except Exception as e:
        logger.error(f'An error occurred. Details: {traceback.format_exc()}')
        if Config.debug:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 500,
                        'msg': f'Unexpected error has occurred [{str(e)}]',
                        'args': {
                            'code': 'CORE_SERVER_ERROR',
                            'err': str(e),
                            'call_id': request['call_id'],
                        },
                    }
                )
            )
        else:
            await env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': 500,
                        'msg': 'Unexpected error has occurred',
                        'args': {
                            'code': 'CORE_SERVER_ERROR',
                            'call_id': request['call_id'],
                        },
                    }
                )
            )


def _call_callback(call_task: 'Future', env: 'NawahEnv', call_id: str):
    try:
        results = call_task.result()
        results['args']['call_id'] = call_id
        asyncio.create_task(env['ws'].send_str(app_encoder.encode(results)))
    except Exception as e:
        status = getattr(e, 'status', 500)
        msg = 'Unexpected error has occurred'
        code = camel_to_upper(e.__class__.__name__)
        if e.args:
            msg = e.args[0]
            if len(e.args) == 2:
                if isinstance(e.args[1], dict):
                    if 'args' in e.args[1]:
                        if 'code' in e.args[1]['args']:
                            code = e.args[1]['args']['code']
                    if 'status' in e.args[1]:
                        status = e.args[1]['status']

        asyncio.create_task(
            env['ws'].send_str(
                app_encoder.encode(
                    {
                        'status': status,
                        'msg': msg,
                        'args': {'code': code, 'call_id': call_id},
                    }
                )
            )
        )



'''Provieds Utilities to be used to get and set cache'''

import copy
import datetime
import inspect
import logging
from typing import TYPE_CHECKING, Any, Optional, Protocol, Tuple, cast

import jwt
import redis.exceptions
from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.classes import Cache, Func, Module, Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger('nawah')

Config.cache_expiry = cast(int, Config.cache_expiry)


class CacheNotConfiguredException(Exception):
    '''raises if a Cache Utility is called while app is not configured for Cache Workflow'''


class UpdateCacheRemoveCondition(Protocol):
    '''Provides type-hint for \'update_cache\' Utility \'remove_condition\' callable'''

    # pylint: disable=too-few-public-methods

    def __call__(self, *, update_doc: 'NawahDoc') -> bool:
        ...


async def check_cache_connection(attempt: int = 3):
    '''Attempts to read from cache to force re-connection if broken'''

    if not Config.sys.cache:
        raise CacheNotConfiguredException()

    try:
        await Config.sys.cache.get('__connection')
    except redis.exceptions.ConnectionError as e:
        if attempt != 0:
            return await check_cache_connection(attempt=attempt - 1)

        raise e


async def reset_cache_channel(channel: str, /):
    '''Resets specific cache `channel` by deleting it from active Redis db'''

    if not (cache := Config.sys.cache):
        raise CacheNotConfiguredException()

    await check_cache_connection()

    try:
        for key in await cache.client.keys(f'{channel}:*'):
            try:
                await cache.client.delete(key.decode('utf-8'))
            except redis.exceptions.ResponseError:
                logger.error('Failed to delete Cache Key: \'%s\'', key)
    except redis.exceptions.ConnectionError:
        logger.error(
            'Connection with Redis server \'%s\' failed. Skipping resetting Cache Channel \'%s\'.',
            Config.cache_server,
            channel,
        )


async def update_cache(
    *,
    channels: list[str],
    docs: list['ObjectId'],
    update_doc: 'NawahDoc',
    remove_condition: 'UpdateCacheRemoveCondition' = None,
):
    if not (cache := Config.sys.cache):
        raise CacheNotConfiguredException()

    remove_key = False
    try:
        if remove_condition:
            remove_key = remove_condition(update_doc=update_doc)
    except:
        remove_key = True

    for channel in channels:

        if channel.endswith(':'):
            channel += '*'
        elif not channel.endswith(':*'):
            channel += ':*'

        try:
            await check_cache_connection()

            for key in await cache.client.keys(channel):
                key = key.decode('utf-8')

                key_docs = await cache.get(key, '.docs')
                for doc in docs:
                    try:
                        doc_index = key_docs.index(str(doc))
                    except ValueError:
                        continue
                    if remove_key:
                        try:
                            await cache.delete(key, '.')
                        except redis.exceptions.ResponseError:
                            logger.error(
                                'Cache command failed with \'ResponseError\'. Current scope:'
                            )
                            logger.error(locals())
                        continue
                    if await cache.get(
                        key, f'.results.args.docs[{doc_index}]._id'
                    ) != str(doc):
                        try:
                            await cache.delete(key, '.')
                        except redis.exceptions.ResponseError as e:
                            logger.error(
                                'Cache command failed with \'ResponseError\': \'%s\'',
                                e,
                            )
                            logger.error('Current scope: %s', locals())
                    try:
                        for attr_name in update_doc:
                            if update_doc[attr_name] == None:
                                continue
                            await cache.set(
                                key,
                                f'.results.args.docs[{doc_index}].{attr_name}',
                                update_doc[attr_name],
                            )
                    except redis.exceptions.ResponseError as e:
                        logger.error(
                            'Cache command failed with \'ResponseError\': \'%s\'',
                            e,
                        )
                        logger.error('Current scope: %s', locals())
                        logger.error(
                            'Removing key \'%s\' due to failed update attempt.', key
                        )
                        try:
                            await cache.delete(key, '.')
                        except redis.exceptions.ResponseError as e:
                            logger.error(
                                'Cache command failed with \'ResponseError\': \'%s\'',
                                e,
                            )
                            logger.error('Current scope: %s', locals())
                try:
                    await cache.set(
                        key,
                        '.results.args.cache_time',
                        datetime.datetime.utcnow().isoformat(),
                    )
                except redis.exceptions.ResponseError as e:
                    logger.error(
                        'Cache command failed with \'ResponseError\': \'%s\'',
                        e,
                    )
                    logger.error('Current scope: %s', locals())
        except redis.exceptions.ConnectionError:
            logger.error(
                'Connection with Redis server \'%s\' failed. Skipping updating Cache Channel \'%s\'.',
                Config.cache_server,
                channel,
            )


def _generate_cache_key(
    *,
    func: 'Func',
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
) -> Optional[str]:
    if not Config.sys.cache or not func.cache:
        return None

    condition_params = {
        'skip_events': skip_events,
        'env': env,
        'query': query,
    }

    if not func.cache.condition(
        **{
            param: condition_params[param]
            for param in inspect.signature(func.cache.condition).parameters
        }
    ):
        return None

    cache_key = {
        'query': _cache_encoder_item('', {}, query._query),
        'special': _cache_encoder('', {}, query._special),
        'user': env['session']['user']['_id'] if func.cache.user_scoped else None,
    }

    cache_key_jwt = jwt.encode(cache_key, '_').split('.')[1]

    return cache_key_jwt


def _cache_encoder(path, files, results):
    for k, v in results.items():
        results[k] = _cache_encoder_item(f'{path}.{k}', files, v)

    return results


def _cache_encoder_item(path, files, item):
    if isinstance(item, list):
        for i, child_item in enumerate(item):
            if isinstance(child_item, dict):
                item[i] = _cache_encoder(f'{path}.{i}', files, child_item)
            else:
                item[i] = _cache_encoder_item(f'{path}.{i}', files, child_item)

        return item

    if isinstance(item, datetime.datetime):
        return item.isoformat()

    if isinstance(item, ObjectId):
        return str(item)

    if isinstance(item, bytes):
        files[path] = item
        return True

    return item


async def _call_cache(func: 'Func', cache_key: str):
    if not Config.sys.cache:
        return

    module = cast('Module', func.module)
    cache = cast('Cache', func.cache)

    try:
        logger.debug(
            'Attempting to get cache with \'key\': \'%s\'.',
            f'.{module.name}.{func.name}.{cache_key}',
        )

        await check_cache_connection()

        if cache.file:
            cache_dict = None
            file_cache = await Config.sys.cache.client.hgetall(
                f'__files__:{cache.channel}:{module.name}:{func.name}:{cache_key}',
            )
            if file_cache:
                cache_dict = {
                    'status': 200,
                    'msg': '',
                    'args': {
                        'return': 'file',
                        'docs': [
                            {
                                file_cache_key.decode('utf-8'): file_cache_val
                                if file_cache_key == b'content'
                                else file_cache_val.decode('utf-8')
                                for file_cache_key, file_cache_val in file_cache.items()
                            }
                        ],
                    },
                }

        else:
            cache_dict = await Config.sys.cache.get(
                f'{cache.channel}:{module.name}:{func.name}:{cache_key}',
                '.results',
            )

        return cache_dict

    except redis.exceptions.ResponseError:
        return

    except redis.exceptions.ConnectionError:
        logger.error(
            'Connection with Redis server \'%s\' failed. Skipping Cache Workflow.',
            Config.cache_server,
        )
        return


async def _set_cache(*, func: 'Func', cache_key: str, results: 'Results'):
    if not Config.sys.cache:
        return

    module = cast('Module', func.module)
    cache = cast('Cache', func.cache)

    cache_key_long = f'{cache.channel}:{module.name}:{func.name}:{cache_key}'
    results = copy.deepcopy(results)

    try:
        logger.debug(
            'Attempting to set cache with \'key\': \'%s\'.',
            cache_key_long,
        )

        await check_cache_connection()

        cache_dict = {
            'docs': [doc['_id'] for doc in results['args']['docs']]
            if 'args' in results and 'docs' in results['args']
            else [],
            'results': results,
        }

        # Check if results contain file, and set using regular Redis cache
        if cache.file:
            # Update value for cache_key_long to prevent conflict with RedisJSON cache
            cache_key_long = f'__files__:{cache_key_long}'
            # Set cache
            await Config.sys.cache.client.hset(
                cache_key_long,
                mapping=results['args']['docs'][0],
            )

        # Otherwise, set as RedisJSON cache
        else:
            await Config.sys.cache.set(
                cache_key_long,
                '.',
                _cache_encoder('', {}, cache_dict),
            )

        if Config.cache_expiry:
            await Config.sys.cache.client.expire(
                cache_key_long,
                Config.cache_expiry,
            )

    except redis.exceptions.ConnectionError:
        logger.error(
            'Connection with Redis server \'%s\' failed. Skipping Cache Workflow.',
            Config.cache_server,
        )


async def _get_cache(
    *,
    func: 'Func',
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
) -> Tuple[Optional[str], Any]:
    cache_key = _generate_cache_key(
        func=func,
        skip_events=skip_events,
        env=env,
        query=query,
    )
    call_cache = None

    if not cache_key:
        return (None, None)

    call_cache = await _call_cache(func=func, cache_key=cache_key)

    return (cache_key, call_cache)



'''Provides 'user' Module Functions callables'''

import copy
from typing import TYPE_CHECKING, Optional

from bson import ObjectId

from nawah.config import Config
from nawah.enums import Event, NawahValues
from nawah.exceptions import FuncException, InvalidAttrException
from nawah.utils import call, encode_attr_type, validate_attr

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _read(
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
    skip_user_settings: Optional[bool],
) -> 'Results':
    skip_events.append(Event.PERM)
    read_results = await call(
        'base/read', module_name='user', skip_events=skip_events, env=env, query=query
    )

    if read_results['status'] != 200:
        return read_results

    if skip_user_settings is True:
        return read_results

    for i, _ in enumerate(read_results['args']['docs']):
        user = read_results['args']['docs'][i]
        for auth_attr in Config.user_attrs:
            del user[f'{auth_attr}_hash']
        if Config.user_doc_settings:
            setting_results = await call(
                'base/read',
                module_name='setting',
                skip_events=[Event.PERM],
                env=env,
                query=[{'user': user['_id'], 'var': {'$in': Config.user_doc_settings}}],
            )
            user_doc_settings = copy.copy(Config.user_doc_settings)
            if setting_results['args']['count']:
                for setting_doc in setting_results['args']['docs']:
                    user_doc_settings.remove(setting_doc['var'])
                    user[setting_doc['var']] = setting_doc['val']
            # Forward-compatibility: If user was created before presence
            # of any user_doc_settings, add them with default value
            for setting_attr in user_doc_settings:
                user[setting_attr] = Config.user_settings[setting_attr].default
                # Set NawahValues.NONE_VALUE to None if it was default
                if user[setting_attr] == NawahValues.NONE_VALUE:
                    user[setting_attr] = None

    return read_results


async def _create(
    skip_events: 'NawahEvents', env: 'NawahEnv', doc: 'NawahDoc'
) -> 'Results':
    if Event.ATTRS_DOC not in skip_events:
        doc['groups'] = [ObjectId('f00000000000000000000013')]
    user_settings = {}
    for setting_name, setting in Config.user_settings.items():
        if setting.type == 'user_sys':
            user_settings[setting_name] = copy.deepcopy(setting.default)
        else:
            if setting_name in doc:
                try:
                    validate_attr(
                        mode='create',
                        attr_name=setting_name,
                        attr_type=setting.val_type,
                        attr_val=doc[setting_name],
                    )
                    user_settings[setting_name] = doc[setting_name]
                except InvalidAttrException as validation_exception:
                    raise FuncException(
                        status=400,
                        msg=f'Invalid settings attr \'{setting_name}\' for \'create\' request on '
                        'module \'user\'',
                        args={'code': 'CREATE_INVALID_SETTING'},
                    ) from validation_exception

            if setting.default == NawahValues.NONE_VALUE:
                raise FuncException(
                    status=400,
                    msg=f'Missing settings attr \'{setting_name}\' for \'create\' request on '
                    'module \'user\'',
                    args={'code': 'MISSING_ATTR'},
                )

            user_settings[setting_name] = copy.deepcopy(setting.default)

    create_results = await call(
        'base/create', skip_events=[Event.PERM], module_name='user', env=env, doc=doc
    )

    for setting_name, setting_val in user_settings.items():
        setting_results = await call(
            'setting/create',
            skip_events=[Event.PERM, Event.ATTRS_DOC],
            env=env,
            doc={
                'user': create_results['args']['docs'][0]['_id'],
                'var': setting_name,
                'val_type': encode_attr_type(
                    attr_type=Config.user_settings[setting_name].val_type
                ),
                'val': setting_val,
                'type': Config.user_settings[setting_name].type,
            },
        )
        if setting_results['status'] != 200:
            return setting_results

    return create_results


async def _read_privileges(env: 'NawahEnv', query: 'Query') -> 'Results':
    # Confirm _id is valid
    results = await call(
        'user/read',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    for group in user['groups']:
        group_results = await call(
            'group/read', skip_events=[Event.PERM], env=env, query=[{'_id': group}]
        )
        group = group_results['args']['docs'][0]
        for privilege in group['privileges'].keys():
            if privilege not in user['privileges'].keys():
                user['privileges'][privilege] = []
            for i in range(len(group['privileges'][privilege])):
                if (
                    group['privileges'][privilege][i]
                    not in user['privileges'][privilege]
                ):
                    user['privileges'][privilege].append(
                        group['privileges'][privilege][i]
                    )
    return results


async def _add_group(
    skip_events: 'NawahEvents', env: 'NawahEnv', query: 'Query', doc: 'NawahDoc'
) -> 'Results':
    # Check for list group attr
    if isinstance(doc['group'], list):
        for i in range(0, len(doc['group']) - 1):
            await call(
                'user/add_group',
                skip_events=skip_events,
                env=env,
                query=query,
                doc={'group': doc['group'][i]},
            )
        doc['group'] = doc['group'][-1]
    # Confirm all basic args are provided
    doc['group'] = ObjectId(doc['group'])
    # Confirm group is valid
    results = await call(
        'group/read', skip_events=[Event.PERM], env=env, query=[{'_id': doc['group']}]
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
        )
    # Get user details
    results = await call('user/read', skip_events=[Event.PERM], env=env, query=query)
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    # Confirm group was not added before
    if doc['group'] in user['groups']:
        raise FuncException(
            status=400,
            msg='User is already a member of the group.',
            args={'code': 'GROUP_ADDED'},
        )
    user['groups'].append(doc['group'])
    # Update the user
    results = await call(
        'user/update',
        skip_events=[Event.PERM],
        env=env,
        query=query,
        doc={'groups': user['groups']},
    )
    # if update fails, return update results
    if results['status'] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env['session']['user']['_id'] == user['_id']:
        user_results = await call(
            'user/read_privileges',
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': user['_id']}],
        )
        env['session']['user'] = user_results['args']['docs'][0]

    return results


async def _delete_group(env: 'NawahEnv', query: 'Query') -> 'Results':
    # Confirm group is valid
    results = await call(
        'group/read',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['group'][0]}],
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
        )
    # Get user details
    results = await call(
        'user/read', skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    # Confirm group was not added before
    if query['group'][0] not in user['groups']:
        raise FuncException(
            status=400,
            msg='User is not a member of the group.',
            args={'code': 'GROUP_NOT_ADDED'},
        )
    # Update the user
    results = await call(
        'user/update',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
        doc={'groups': {'$del_val': [query['group'][0]]}},
    )
    # if update fails, return update results
    if results['status'] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env['session']['user']['_id'] == user['_id']:
        user_results = await call(
            'user/read_privileges',
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': user['_id']}],
        )
        env['session']['user'] = user_results['args']['docs'][0]

    return results



'''Provides enums required for Nawah Fremework and apps'''

from ._enums import (AttrType, DeleteStrategy, Event, LocaleStrategy,
                     NawahValues, VarType)

__all__ = [
    'AttrType',
    'DeleteStrategy',
    'Event',
    'LocaleStrategy',
    'NawahValues',
    'VarType',
]



'''Provides types used in Nawah'''

from typing import (TYPE_CHECKING, Any, Literal, MutableMapping,
                    MutableSequence, Optional, TypedDict, Union)

from nawah.enums import Event

if TYPE_CHECKING:
    from datetime import datetime

    from aiohttp.web import WebSocketResponse
    from motor.motor_asyncio import AsyncIOMotorClient

    from .._package import Package


class Results(TypedDict):
    '''Provides type-hint for return of Nawah Function'''

    status: int
    msg: str
    args: 'ResultsArgs'


ResultsArgs = MutableMapping[str, Any]


class AppPackage(TypedDict):
    '''Provides type-hint for item of 'Config.packages' '''

    package: 'Package'
    modules: MutableSequence[str]


class AnalyticsEvents(TypedDict):
    '''Provides type-hint for 'analytics_events' App Config Attr'''

    app_conn_verified: bool
    session_conn_auth: bool
    session_user_auth: bool
    session_conn_reauth: bool
    session_user_reauth: bool
    session_conn_deauth: bool
    session_user_deauth: bool


class IPQuota(TypedDict):
    '''Provides type-hint for dict used to track IP user quota'''

    counter: int
    last_check: 'datetime'


NawahEvents = MutableSequence[Event]


class NawahEnv(TypedDict, total=False):
    '''Provides type-hint for 'Env' dict'''

    id: str
    init: bool
    conn: 'AsyncIOMotorClient'
    REMOTE_ADDR: str
    HTTP_USER_AGENT: str
    HTTP_ORIGIN: str
    client_app: str
    session: 'NawahSession'
    prev_session: MutableMapping
    prev_session_timeout: 'datetime'
    last_call: 'datetime'
    ws: 'WebSocketResponse'
    quota: 'NawahEnvQuota'
    args: MutableMapping[str, Any]


class NawahEnvQuota(TypedDict):
    '''Provides type-hint for 'quota' of 'Env' dict'''

    counter: int
    last_check: 'datetime'


class NawahSession(TypedDict):
    '''Provieds type-hint for 'session' of 'Env' dict'''

    user: MutableMapping
    groups: MutableSequence[str]
    host_add: str
    user_agent: str
    expiry: str
    token_hash: str
    create_time: str


NawahQuery = MutableSequence[  # type: ignore
    Union[
        'NawahQuery',  # type: ignore
        Union[
            MutableMapping[
                str,
                Union[
                    'NawahQuery',  # type: ignore
                    Any,
                    Union[
                        MutableMapping[Literal['$eq'], Any],
                        MutableMapping[Literal['$ne'], Any],
                        MutableMapping[Literal['$gt'], Union[int, str]],
                        MutableMapping[Literal['$gte'], Union[int, str]],
                        MutableMapping[Literal['$lt'], Union[int, str]],
                        MutableMapping[Literal['$lte'], Union[int, str]],
                        MutableMapping[Literal['$all'], MutableSequence[Any]],
                        MutableMapping[Literal['$in'], MutableSequence[Any]],
                        MutableMapping[Literal['$nin'], MutableSequence[Any]],
                        MutableMapping[Literal['$regex'], str],
                    ],
                ],
            ],
            'NawahQuerySpecial',
        ],
    ]
]


class NawahQuerySpecialGroup(TypedDict):
    '''Provides type-hint for '$group' in 'NawahQuery' '''

    by: str
    count: int


class NawahQuerySpecialGeoNear(TypedDict):
    '''Provides type-hint for '$geo_near' in 'NawahQuery' '''

    val: str
    attr: str
    dist: int


# Following TypedDict type can't be defined as class as keys include $
NawahQuerySpecial = TypedDict(
    'NawahQuerySpecial',
    {
        '$search': Optional[str],
        '$sort': Optional[MutableMapping[str, Literal[1, -1]]],
        '$skip': Optional[int],
        '$limit': Optional[int],
        '$extn': Optional[Union[bool, MutableSequence[str]]],
        '$attrs': Optional[MutableSequence[str]],
        '$group': Optional[MutableSequence[NawahQuerySpecialGeoNear]],
        '$geo_near': Optional[NawahQuerySpecialGeoNear],
    },
    total=False,
)

NawahDoc = MutableMapping[str, Any]



'''Provides classes used in Nawah'''

from ._analytics import Analytics
from ._attr import Attr
from ._cache import Cache
from ._client_app import ClientApp
from ._counter import Counter
from ._default import Default
from ._diff import Diff
from ._encoders import app_encoder
from ._extn import Extn
from ._func import Func, Perm
from ._job import Job
from ._l10n import L10N
from ._module import Module
from ._package import App, Env, Package
from ._query import Query
from ._sys_doc import SysDoc
from ._user_setting import UserSetting
from ._var import Var

__all__ = [
    'Analytics',
    'Attr',
    'Cache',
    'ClientApp',
    'Counter',
    'Default',
    'Diff',
    'app_encoder',
    'Extn',
    'Func',
    'Perm',
    'Job',
    'L10N',
    'Module',
    'App',
    'Env',
    'Package',
    'Query',
    'SysDoc',
    'UserSetting',
    'Var',
]



'''Provides utilities required for configuring modules, packages and app'''

import datetime
import logging
import os
from typing import TYPE_CHECKING, Any, MutableMapping, cast

import sentry_sdk
from bson import ObjectId
from croniter import croniter
from passlib.hash import pbkdf2_sha512
from redis import asyncio as aioredis  # type: ignore

from nawah.classes import (Attr, Counter, Default, Diff, Extn, Func, Module,
                           SysDoc, UserSetting)
from nawah.config import Config
from nawah.enums import AttrType, Event, NawahValues
from nawah.exceptions import (ConfigException, InvalidAttrTypeArgException,
                              InvalidAttrTypeException)

from ._attr_type import extract_attr, generate_attr_val
from ._call import call
from ._validate._type import validate_type
from ._validate._user_setting import validate_user_setting

if TYPE_CHECKING:
    from nawah.types import NawahDoc

logger = logging.getLogger('nawah')


def config_module(*, module_name: str, module: 'Module'):
    '''Validates args for 'module' and updates values for 'attrs' of 'module' per sets from
    'defaults', 'extns', 'counters' '''

    # pylint: disable=import-outside-toplevel

    import nawah.base as Base

    # Set name of module in Module object for reverse reference
    module.name = module_name

    logger.debug('Attempting to configure module \'%s\'', module_name)

    if not module.funcs:
        raise ConfigException(f'Module \'{module_name}\' doesn\'t define any function')

    for func_name, func in module.funcs.items():
        logger.debug(
            'Attempting to validate Function \'%s\'.\'%s\'', module_name, func_name
        )
        func.name = func_name
        func.module = module
        # Check value type
        if not isinstance(func, Func):
            raise ConfigException(
                f'Invalid Nawah Function for \'{module_name}\'.\'{func_name}\' of type '
                '\'{type(func)}\''
            )

        # Validate query_attrs, doc_attrs, call_args
        for func_attr_name in ['query_attrs', 'doc_attrs', 'call_args']:
            logger.debug(
                'Attempting to validate \'%s\' of \'%s\'.\'%s\'',
                func_attr_name,
                module_name,
                func_name,
            )
            if func_attr_val := getattr(func, func_attr_name):
                # query_attrs, doc_attrs are lists of dicts, call_args is just dict
                # Do this trick to get this loop to work nonetheless, by wrapping call_args with
                # a list, to let the rest of the loop deal with theree attrs in same manner
                if func_attr_name == 'call_args':
                    func_attr_val = [func_attr_val]
                for attrs_set in func_attr_val:
                    for attr_name, attr_type in attrs_set.items():
                        try:
                            logger.debug(
                                'Attempting to validate \'%s\' Attr Type of \'%s\' for Function '
                                '\'%s\'.\'%s\' of set \'%s\'',
                                func_attr_name,
                                attr_name,
                                module_name,
                                func.name,
                                attrs_set,
                            )
                            validate_type(attr_type=attr_type, skip_type=True)
                        except InvalidAttrTypeException as e:
                            raise ConfigException(
                                f'Invalid Attr Type for \'{module_name}\'.\'{attr_name}\'. '
                                f'Original validation error: {str(e)}',
                            ) from e
                        except InvalidAttrTypeArgException as e:
                            raise ConfigException(
                                f'Invalid Attr Type Arg for \'{module_name}\'.\'{attr_name}\'. '
                                f'Original validation error: {str(e)}'
                            ) from e

        # [TODO] Add steps to update Func object attributes per func callable
        # If Func has callable defined skip
        if func.callable:
            if not callable(func.callable):
                raise ConfigException(
                    f'Value \'callable\' for Nawah Function \'{func_name}\' of \'{module_name}\' '
                    'is not callable'
                )
            continue

        logger.debug(
            'Function \'%s\'.\'%s\' doesn\'t define callable. Attempting to set per name',
            module_name,
            func_name,
        )
        # Otherwise, set callable based on fucntion name, update query_attrs, doc_attrs, call_args
        if func_name.startswith('create_file'):
            func.doc_attrs = [
                {
                    'attr': Attr.STR(),
                    'lastModified': Attr.INT(),
                    'type': Attr.STR(),
                    'name': Attr.STR(),
                    'content': Attr(
                        desc='',
                        type=AttrType.BYTES,
                        args={},
                    ),
                },
                {'_id': Attr.ID(), 'doc': Attr.ID()},
            ]
            func.post_func = True
            func.callable = Base.create_file

        elif func_name.startswith('update_file'):
            func.post_func = True
            func.callable = Base.update_file

        elif func_name.startswith('delete_file'):
            func.post_func = True
            func.callable = Base.delete_file

        elif func_name.startswith('delete_lock'):
            func.callable = Base.delete_lock

        elif func_name.startswith('obtain_lock'):
            func.callable = Base.obtain_lock

        elif func_name.startswith('read_diff'):
            func.callable = Base.read_diff

        elif func_name.startswith('retrieve_file'):
            func.query_attrs = [
                {
                    '_id': Attr.ID(),
                    'attr': Attr.STR(),
                    'filename': Attr.STR(),
                },
                {
                    '_id': Attr.ID(),
                    'attr': Attr.STR(),
                    'thumb': Attr.STR(pattern=r'[0-9]+x[0-9]+'),
                    'filename': Attr.STR(),
                },
            ]
            func.get_func = True
            func.callable = Base.retrieve_file

        elif func_name.startswith('read'):
            func.callable = Base.read

        elif func_name.startswith('create'):
            func.callable = Base.create

        elif func_name.startswith('update'):
            func.callable = Base.update

        elif func_name.startswith('delete'):
            func.callable = Base.delete

        else:
            raise ConfigException(
                f'Can\'t set callable for Nawah Function \'{func_name}\' of'
                f' \'{module_name}\''
            )

        if not func.call_args:
            func.call_args = {}

        func.call_args.update({'raise_no_success': Attr.BOOL()})

    # Check attrs for any invalid type
    for attr in module.attrs:
        try:
            logger.debug(
                'Attempting to validate Attr Type for \'%s\'.\'%s\'',
                module_name,
                attr,
            )
            validate_type(attr_type=module.attrs[attr], skip_type=True)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f'Invalid Attr Type for \'{module_name}\'.\'{attr}\'. Original '
                f'validation error: {str(e)}',
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f'Invalid Attr Type Arg for \'{module_name}\'.\'{attr}\'. Original '
                f'validation error: {str(e)}'
            ) from e

    # Check defaults for invalid types, update default value
    for attr_name, default in module.defaults.items():
        if not isinstance(default, Default):
            raise ConfigException(
                f'Invalid Default Instruction for \'{module_name}\'.\'{attr_name}\', of type '
                f'\'{type(default)}\''
            )

        logger.debug(
            'Updating default value for attr \'%s\' to: \'%s\'',
            attr_name,
            module.defaults[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).default = module.defaults[
            attr_name
        ]

    # Check counters for invalid attrs, types
    for attr_name, counter in module.counters.items():
        if not isinstance(counter, Counter):
            raise ConfigException(
                f'Invalid Counter Instruction for \'{module_name}\'.\'{attr_name}\', of type '
                f'\'{type(counter)}\''
            )

        if attr_name not in module.attrs:
            raise ConfigException(
                f'Invalid attr for Counter Instruction. Module \'{module_name}\' '
                f'doesn\'t have top-level attr \'{attr_name}\''
            )

        if module.attrs[attr_name].type != AttrType.ANY:
            raise ConfigException(
                f'Attr \'{module_name}\'.\'{attr_name}\' for Counter Instruction can only be of '
                'Attr Type ANY'
            )

        # Set default value to Default(value=None) to skip InvalidAttrException when checking doc
        module.attrs[attr_name].default = Default(value=None)

    # Update extn value
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f'Invalid Extension Instruction for \'{module_name}\'.\'{attr_name}\', of type '
                f'\'{type(extn)}\''
            )

        logger.debug(
            'Updating \'extn\' for attr \'%s\' to: \'%s\'',
            attr_name,
            module.extns[attr_name],
        )
        extract_attr(attrs=module.attrs, path=attr_name).extn = module.extns[attr_name]

    # Check extns for invalid extended attrs
    for attr_name, extn in module.extns.items():
        if not isinstance(extn, Extn):
            raise ConfigException(
                f'Invalid extns attr \'{attr_name}\' of module \'{module_name}\''
            )

    # Check valid type, value for diff
    if module.diff is not None and not isinstance(module.diff, Diff):
        raise ConfigException(f'Invalid Diff Instruction for module \'{module_name}\'')

    if isinstance(module.diff, Attr):
        module.diff = cast(Attr, module.diff)
        if module.diff.type != AttrType.TYPE:
            raise ConfigException(
                f'Invalid Attr Type for diff of module \'{module_name}\'. Only Attr Type TYPE '
                'is allowed'
            )
        logger.debug(
            'Attempting to validate Attr Type diff of module \'%s\'', module_name
        )
        try:
            validate_type(attr_type=module.diff)
        except InvalidAttrTypeException as e:
            raise ConfigException(
                f'Invalid Attr Type for diff of module \'{module_name}\'. Original validation error: {str(e)}'
            ) from e
        except InvalidAttrTypeArgException as e:
            raise ConfigException(
                f'Invalid Attr Type Arg for diff of module \'{module_name}\'. Original validation error: {str(e)}'
            ) from e

    # [TODO] Re-implement
    # Check valid types, values for create_draft, update_draft
    # for attr in ['create_draft', 'update_draft']:
    #     if type(getattr(self, attr)) not in [bool, Attr]:
    #         raise ConfigException(f'Invalid {attr} for module \'{module_name}\'') from e
    #     if type(getattr(self, attr)) == Attr:
    #         if getattr(self, attr).type != AttrType.TYPE:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Only Attr Type TYPE is allowed'
    #             ) from e
    #         logger.debug(
    #             f'Attempting to validate Attr Type {attr} of module \'{module_name}\'.'
    #         )
    #         try:
    #             validate_type(attr_type=getattr(self, attr))
    #         except InvalidAttrTypeException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e
    #         except InvalidAttrTypeArgException as e:
    #             raise ConfigException(
    #                 f'Invalid Attr Type Arg for {attr} of module \'{module_name}\'. Original validation error: {str(e)}'
    #             ) from e

    logger.debug('Configured module %s', module_name)


async def config_app():

    # [TODO] Re-implement
    # Check API version
    # nawah_level = '.'.join(Config.sys.version.split('.')[0:2])
    # for package, api_level in {
    #     name: config['package'].api_level
    #     for name, config in Config.sys.packages.items()
    # }.items():
    #     if api_level != nawah_level:
    #         raise ConfigException(
    #             f'Nawah framework is on API-level \'{nawah_level}\', but the app package'
    #             f'\'{package}\' requires API-level \'{api_level}\''
    #         )
    # try:
    #     versions = (
    #         (
    #             requests.get(
    #                 'https://raw.githubusercontent.com/masaar/nawah_versions/master/versions.txt'
    #             ).content
    #         )
    #         .decode('utf-8')
    #         .split('\n')
    #     )
    #     version_detected = ''
    #     for version in versions:
    #         if version.startswith(f'{nawah_level}.'):
    #             if version_detected and int(version.split('.')[-1]) < int(
    #                 version_detected.split('.')[-1]
    #             ):
    #                 continue
    #             version_detected = version
    #     if version_detected and version_detected != Config.sys.version:
    #         logger.warning(
    #             f'Your app is using Nawah version \'{Config.sys.version}\' while newer version \'{version_detected}\' of the API-level is available. Please, update'
    #         )
    # except:
    #     logger.warning(
    #         'An error occurred while attempting to check for latest update to Nawah. Please, check for updates on your own'
    #     )

    # Check for jobs
    if Config.jobs:
        # Check jobs schedule validity
        Config.sys.jobs_base = datetime.datetime.utcnow()
        for job_name, job in Config.jobs.items():
            if not croniter.is_valid(job.schedule):
                raise ConfigException(
                    f'Job with schedule \'{job_name}\' schedule is invalid'
                )

            job.cron_schedule = croniter(job.schedule, Config.sys.jobs_base)
            job.next_time = datetime.datetime.fromtimestamp(
                job.cron_schedule.get_next(), datetime.timezone.utc
            ).isoformat()[:16]

    # Check for presence of user_auth_attrs
    if not Config.user_attrs:
        raise ConfigException('No \'user_attrs\' are provided')

    # Check for Env Vars
    attrs_defaults = {
        'data_server': 'mongodb://localhost',
        'data_name': 'nawah_data',
        'data_ssl': False,
        'emulate_test': False,
    }
    for attr_name in attrs_defaults:
        attr_val = getattr(Config, attr_name)
        if type(attr_val) == str and attr_val.startswith('$__env.'):
            logger.debug(f'Detected Env Variable for config attr \'{attr_name}\'')
            if not os.getenv(attr_val[7:]):
                logger.warning(
                    f'Couldn\'t read Env Variable for config attr \'{attr_name}\'. Defaulting to \'{attrs_defaults[attr_name]}\''
                )
                setattr(Config, attr_name, attrs_defaults[attr_name])
            else:
                # Set data_ssl to True rather than string Env Variable value
                if attr_name == 'ssl':
                    attr_val = True
                else:
                    attr_val = os.getenv(attr_val[7:])
                logger.warning(
                    f'Setting Env Variable for config attr \'{attr_name}\' to \'{attr_val}\''
                )
                setattr(Config, attr_name, attr_val)

    from nawah import data as Data

    # Create default env dict
    anon_user = _compile_anon_user()
    anon_session = _compile_anon_session()
    anon_session['user'] = anon_user
    Config.sys.conn = Data.create_conn()
    Config.sys.env = {
        'conn': Config.sys.conn,
        'REMOTE_ADDR': '127.0.0.1',
        'HTTP_USER_AGENT': 'Nawah',
        'client_app': '__sys',
        'session': anon_session,
    }

    # Check cache_server Config Attr
    if Config.cache_server:
        logger.debug('Attempting to create connection with \'cache_server\'')
        Config.sys.cache = aioredis.Redis(
            host=Config.cache_server,
            db=Config.cache_db,
            username=Config.cache_username,
            password=Config.cache_password,
        ).json()

    # Check error_reporting_server Config Attr
    if Config.error_reporting_server:
        logger.debug('Attempting to create connection with \'error_reporting_server\'')
        sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
            Config.error_reporting_server,
            traces_sample_rate=1.0,
            with_locals=True,
            release=f'{Config.sys.name}:{Config.sys.version}',
            environment=Config.env,
        )

    # Check test mode
    if Config.test:
        logger.debug('Test mode detected')
        logger.setLevel(logging.DEBUG)
        __location__ = os.path.realpath(os.path.join('.'))
        if not os.path.exists(os.path.join(__location__, 'tests')):
            os.makedirs(os.path.join(__location__, 'tests'))
        for module_name, module in Config.modules.items():
            module_collection = module.collection
            logger.debug(
                f'Updating collection name \'{module_collection}\' of module {module_name}'
            )
            module_collection = module.collection = f'test_{module_collection}'

            logger.debug(f'Flushing test collection \'{module_collection}\'')
            await Data.drop(
                env=Config.sys.env,
                collection_name=module_collection,
            )

    # Test user_settings
    logger.debug('Testing user_settings')
    if Config.user_settings:
        for user_setting in Config.user_settings:
            logger.debug('Testing %s', user_setting)
            if not isinstance(Config.user_settings[user_setting], UserSetting):
                raise ConfigException(
                    'Invalid Config Attr \'user_settings\' with key \'{user_setting}\' of type \'{type(Config.user_settings[user_setting])}\' with required type \'USER_SETTING\''
                )

            # Validate USER_SETTING
            validate_user_setting(user_setting=Config.user_settings[user_setting])

    # Checking users collection
    # [TODO] Updated sequence to handle users
    logger.debug('Testing users collection')
    if Config.sys.admin_check:
        user_results = await call(
            'user/read',
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            env=Config.sys.env,
            query=[{'_id': 'f00000000000000000000010'}],
        )
        if not user_results['args']['count']:
            logger.debug('ADMIN user not found, creating it')
            # Prepare base ADMIN user doc
            admin_create_doc = {
                '_id': ObjectId('f00000000000000000000010'),
                'name': {Config.locale: '__ADMIN'},
                'groups': [],
                'privileges': {'*': ['*']},
                'locale': Config.locale,
            }
            # Update ADMIN user doc with admin_doc Config Attr
            admin_create_doc.update(Config.admin_doc)

            for auth_attr in Config.user_attrs:
                admin_create_doc[f'{auth_attr}_hash'] = pbkdf2_sha512.using(
                    rounds=100000
                ).hash(
                    f'{auth_attr}{admin_create_doc[auth_attr]}{Config.admin_password}'
                    f'{Config.anon_token}'.encode('utf-8')
                )
            admin_results = await call(
                'user/create',
                skip_events=[Event.PERM],
                env=Config.sys.env,
                doc=admin_create_doc,
            )
            logger.debug('ADMIN user creation results: %s', admin_results)
            if admin_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug(
                '\'ADMIN\' doc found. Attempting to check if update is required..'
            )
            admin_doc = user_results['args']['docs'][0]
            admin_doc_update = {}
            for attr_name, attr_val in Config.admin_doc.items():
                if (
                    attr_name not in admin_doc
                    or not admin_doc[attr_name]
                    or attr_val != admin_doc[attr_name]
                ):
                    if (
                        isinstance(attr_val, dict)
                        and Config.locale in attr_val
                        and isinstance(admin_doc[attr_name], dict)
                        and (
                            (
                                Config.locale in admin_doc[attr_name]
                                and attr_val[Config.locale]
                                == admin_doc[attr_name][Config.locale]
                            )
                            or (Config.locale not in admin_doc[attr_name])
                        )
                    ):
                        continue
                    logger.debug(
                        'Detected change in \'admin_doc.%s\' Config Attr', attr_name
                    )
                    admin_doc_update[attr_name] = attr_val
            for auth_attr in Config.user_attrs:
                auth_attr_hash = pbkdf2_sha512.using(rounds=100000).hash(
                    f'{auth_attr}{admin_doc[auth_attr]}{Config.admin_password}'
                    f'{Config.anon_token}'.encode('utf-8')
                )
                if (
                    f'{auth_attr}_hash' not in admin_doc
                    or auth_attr_hash != admin_doc[f'{auth_attr}_hash']
                ):
                    logger.debug(f'Detected change in \'admin_password\' Config Attr')
                    admin_doc_update[f'{auth_attr}_hash'] = auth_attr_hash
            if len(admin_doc_update.keys()):
                logger.debug(
                    f'Attempting to update ADMIN user with doc: \'{admin_doc_update}\''
                )
                admin_results = await call(
                    'user/update',
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    env=Config.sys.env,
                    query=[{'_id': ObjectId('f00000000000000000000010')}],
                    doc=admin_doc_update,
                )
                logger.debug(f'ADMIN user update results: {admin_results}')
                if admin_results['status'] != 200:
                    raise ConfigException('Config step failed')
            else:
                logger.debug('ADMIN user is up-to-date')
    else:
        logger.warning(
            'Skipping checking \'ADMIN\' doc due to \'sys.admin_check\' Config Attr'
        )

    Config.sys.docs[ObjectId('f00000000000000000000010')] = SysDoc(module='user')

    # Test if ANON user exists
    user_results = await call(
        'user/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000011'}],
    )
    if not user_results['args']['count']:
        logger.debug('ANON user not found, creating it')
        anon_results = await call(
            'user/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_user(),
        )
        logger.debug(f'ANON user creation results: {anon_results}')
        if anon_results['status'] != 200:
            raise ConfigException('Config step failed')
    else:
        logger.debug('ANON user found, checking it')
        anon_doc = _compile_anon_user()
        anon_doc_update = {}
        for attr in Config.user_attrs:
            if attr not in anon_doc or not anon_doc[attr]:
                logger.debug(f'Detected change in \'anon_doc.{attr}\' Config Attr')
                anon_doc_update[attr] = generate_attr_val(
                    attr_type=Config.user_attrs[attr]
                )
        for module in Config.anon_privileges:
            if module not in anon_doc or set(anon_doc[module]) != set(
                Config.anon_privileges[module]
            ):
                logger.debug(f'Detected change in \'anon_privileges\' Config Attr')
                anon_doc_update[f'privileges.{module}'] = Config.anon_privileges[module]
        for auth_attr in Config.user_attrs:
            if (
                f'{auth_attr}_hash' not in anon_doc
                or anon_doc[f'{auth_attr}_hash'] != Config.anon_token
            ):
                logger.debug(f'Detected change in \'anon_token\' Config Attr')
                anon_doc_update[attr] = Config.anon_token
            anon_doc_update[f'{auth_attr}_hash'] = Config.anon_token
        if len(anon_doc_update.keys()):
            logger.debug(
                f'Attempting to update ANON user with doc: \'{anon_doc_update}\''
            )
            anon_results = await call(
                'user/update',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{'_id': ObjectId('f00000000000000000000011')}],
                doc=anon_doc_update,
            )
            logger.debug(f'ANON user update results: {anon_results}')
            if anon_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug('ANON user is up-to-date')

    Config.sys.docs[ObjectId('f00000000000000000000011')] = SysDoc(module='user')

    logger.debug('Testing sessions collection')
    # Test if ANON session exists
    session_results = await call(
        'session/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000012'}],
    )
    if not session_results['args']['count']:
        logger.debug('ANON session not found, creating it')
        anon_results = await call(
            'session/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=_compile_anon_session(),
        )
        logger.debug(f'ANON session creation results: {anon_results}')
        if anon_results['status'] != 200:
            raise ConfigException('Config step failed')
    Config.sys.docs[ObjectId('f00000000000000000000012')] = SysDoc(module='session')

    logger.debug('Testing groups collection')
    # Test if DEFAULT group exists
    group_results = await call(
        'group/read',
        # [TODO] Check behaviour with removed Event.ON
        skip_events=[Event.PERM],
        env=Config.sys.env,
        query=[{'_id': 'f00000000000000000000013'}],
    )
    if not group_results['args']['count']:
        logger.debug('DEFAULT group not found, creating it')
        group_create_doc = {
            '_id': ObjectId('f00000000000000000000013'),
            'user': ObjectId('f00000000000000000000010'),
            'name': {locale: '__DEFAULT' for locale in Config.locales},
            'bio': {locale: '__DEFAULT' for locale in Config.locales},
            'privileges': Config.default_privileges,
        }
        group_results = await call(
            'group/create',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM],
            env=Config.sys.env,
            doc=group_create_doc,
        )
        logger.debug('DEFAULT group creation results: %s', group_results)
        if group_results['status'] != 200:
            raise ConfigException('Config step failed')
    else:
        logger.debug('DEFAULT group found, checking it')
        default_doc = group_results['args']['docs'][0]
        default_doc_update: MutableMapping[str, Any] = {}
        for module in Config.default_privileges:
            if module not in default_doc['privileges'].keys() or set(
                default_doc['privileges'][module]
            ) != set(Config.default_privileges[module]):
                logger.debug('Detected change in \'default_privileges\' Config Attr')
                default_doc_update[f'privileges.{module}'] = Config.default_privileges[
                    module
                ]
        if len(default_doc_update.keys()):
            logger.debug(
                'Attempting to update DEFAULT group with doc: \'%s\'',
                default_doc_update,
            )
            default_results = await call(
                'group/update',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                query=[{'_id': ObjectId('f00000000000000000000013')}],
                doc=default_doc_update,
            )
            logger.debug(f'DEFAULT group update results: {default_results}')
            if anon_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug('DEFAULT group is up-to-date')

    Config.sys.docs[ObjectId('f00000000000000000000013')] = SysDoc(module='group')

    # Test app-specific groups
    logger.debug('Testing app-specific groups collection')
    for group in Config.groups:
        group_results = await call(
            'group/read',
            # [TODO] Check behaviour with removed Event.ON
            skip_events=[Event.PERM],
            env=Config.sys.env,
            query=[{'_id': group['_id']}],
        )
        if not group_results['args']['count']:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' not found, creating it'
            )
            group_results = await call(
                'group/create',
                # [TODO] Check behaviour with removed Event.ON
                # [TODO] Check behaviour with removed Event.PRE
                skip_events=[Event.PERM],
                env=Config.sys.env,
                doc=group,
            )
            logger.debug(
                f'App-specific group with name {group["name"]} creation results: {group_results}'
            )
            if group_results['status'] != 200:
                raise ConfigException('Config step failed')
        else:
            logger.debug(
                f'App-specific group with name \'{group["name"]}\' found, checking it'
            )
            group_doc = group_results['args']['docs'][0]
            group_doc_update = {}
            if 'privileges' in group:
                for module in group['privileges']:
                    if module not in group_doc['privileges'].keys() or set(
                        group_doc['privileges'][module]
                    ) != set(group['privileges'][module]):
                        logger.debug(
                            f'Detected change in \'privileges\' Doc Arg for group with name \'{group["name"]}\''
                        )
                        group_doc_update[f'privileges.{module}'] = group['privileges'][
                            module
                        ]
            if len(group_doc_update.keys()):
                logger.debug(
                    'Attempting to update group with name \'%s\' with doc: \'%s\'',
                    group['name'],
                    group_doc_update,
                )
                group_results = await call(
                    'group/update',
                    # [TODO] Check behaviour with removed Event.ON
                    # [TODO] Check behaviour with removed Event.PRE
                    skip_events=[Event.PERM],
                    env=Config.sys.env,
                    query=[{'_id': group['_id']}],
                    doc=group_doc_update,
                )
                logger.debug(
                    'Group with name \'%s\' update results: %s',
                    group['name'],
                    group_results,
                )
                if group_results['status'] != 200:
                    raise ConfigException('Config step failed')
            else:
                logger.debug('Group with name \'%s\' is up-to-date', group['name'])

        Config.sys.docs[ObjectId(group['_id'])] = SysDoc(module='group')

    # Test app-specific data indexes
    logger.debug('Testing data indexes')
    for index in Config.data_indexes:
        logger.debug('Attempting to create data index: %s', index)
        try:
            Config.sys.conn[Config.data_name][index['collection']].create_index(
                index['index']
            )
        except Exception as e:
            logger.error('Failed to create data index: %s, with error: %s', index, e)
            logger.error('Evaluate error and take action manually')

    logger.debug(
        'Creating \'var\', \'type\', \'user\' data indexes for settings collections'
    )
    Config.sys.conn[Config.data_name]['settings'].create_index([('var', 1)])
    Config.sys.conn[Config.data_name]['settings'].create_index([('type', 1)])
    Config.sys.conn[Config.data_name]['settings'].create_index([('user', 1)])
    logger.debug(
        'Creating \'user\', \'event\', \'subevent\' data indexes for analytics collections'
    )
    Config.sys.conn[Config.data_name]['analytics'].create_index([('user', 1)])
    Config.sys.conn[Config.data_name]['analytics'].create_index([('event', 1)])
    Config.sys.conn[Config.data_name]['analytics'].create_index([('subevent', 1)])
    logger.debug('Creating \'__deleted\' data indexes for all collections')
    for module in Config.modules:
        if Config.modules[module].collection:
            logger.debug(
                'Attempting to create \'__deleted\' data index for collection: %s',
                Config.modules[module].collection,
            )
            Config.sys.conn[Config.data_name][
                Config.modules[module].collection
            ].create_index([('__deleted', 1)])

    # Test app-specific docs
    logger.debug('Testing docs')
    for doc in Config.docs:
        if not isinstance(doc, SysDoc):
            raise ConfigException('Invalid Config Attr \'docs\'')

        doc_results = await call(
            f'{doc.module}/read',
            # [TODO] Check behaviour with removed Event.ON
            # [TODO] Check behaviour with removed Event.PRE
            skip_events=[Event.PERM, Event.ATTRS_QUERY],
            env=Config.sys.env,
            query=[{doc.key: doc.key_value}],  # type: ignore
        )
        if not doc_results['args']['count']:
            skip_events = [Event.PERM]
            if doc.skip_args is True:
                skip_events.append(Event.ATTRS_DOC)
            doc.doc = cast('NawahDoc', doc.doc)
            doc_results = await call(
                f'{doc.module}/create',
                skip_events=skip_events,
                env=Config.sys.env,
                doc=doc.doc,
            )
            logger.debug(
                'App-specific doc with %s \'%s\' of module \'%s\' creation results: %s',
                doc.key,
                doc.key_value,
                doc.module,
                doc_results,
            )
            if doc_results['status'] != 200:
                raise ConfigException('Config step failed')
        Config.sys.docs[ObjectId(doc_results['args']['docs'][0]['_id'])] = SysDoc(
            module=doc.module
        )

    # Check for emulate_test mode
    if Config.emulate_test:
        Config.test = True


def _compile_anon_user():
    anon_doc = {
        '_id': ObjectId('f00000000000000000000011'),
        'name': {Config.locale: '__ANON'},
        'groups': [],
        'privileges': Config.anon_privileges,
        'locale': Config.locale,
    }
    for attr_name, attr_val in Config.user_attrs.items():
        anon_doc[attr_name] = generate_attr_val(attr_type=attr_val)
    for auth_attr in Config.user_attrs:
        anon_doc[f'{auth_attr}_hash'] = Config.anon_token
    for attr_name, user_setting in Config.user_settings.items():
        if user_setting.default is not NawahValues.NONE_VALUE:
            anon_doc[attr_name] = user_setting.default
    if Config.anon_doc:
        anon_doc.update(Config.anon_doc)
    return anon_doc


def _compile_anon_session():
    session_doc = {
        '_id': ObjectId('f00000000000000000000012'),
        'user': ObjectId('f00000000000000000000011'),
        'host_add': '127.0.0.1',
        'user_agent': Config.anon_token,
        'timestamp': '1970-01-01T00:00:00',
        'expiry': '1970-01-01T00:00:00',
        'token': Config.anon_token,
        'token_hash': Config.anon_token,
    }
    return session_doc



'''Provides exceptions related to Query'''

from typing import Any, Literal


class InvalidQueryArgException(Exception):
    '''Raised by \'Query.validate_arg\' if an invalid \'Query Arg\' is detected'''

    status = 400

    def __init__(
        self,
        *,
        arg_name: str,
        arg_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
        arg_type: Any,
        arg_val: Any,
    ):
        super().__init__(
            InvalidQueryArgException.format_msg(
                arg_name=arg_name, arg_oper=arg_oper, arg_type=arg_type, arg_val=arg_val
            ),
            {
                'arg_name': arg_name,
                'arg_oper': arg_oper,
                'arg_type': arg_type,
                'arg_val': arg_val,
            },
        )

    @staticmethod
    def format_msg(*, arg_name, arg_oper, arg_type, arg_val):
        '''Formats exception message'''

        return f'Invalid value for Query Arg \'{arg_name}\' with Query Arg Oper \'{arg_oper}\' expecting type \'{arg_type}\' but got \'{arg_val}\''


class UnknownQueryArgException(Exception):
    '''Raised by \'Query.validate_arg\' if an unknown \'Query Arg\' is detected'''

    status = 400

    def __init__(
        self,
        *,
        arg_name: str,
        arg_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
    ):
        super().__init__(
            UnknownQueryArgException.format_msg(arg_name=arg_name, arg_oper=arg_oper),
            {'arg_name': arg_name, 'arg_oper': arg_oper},
        )

    @staticmethod
    def format_msg(*, arg_name, arg_oper):
        '''Formats exception message'''

        return f'Unknown Query Arg Oper \'{arg_oper}\' for Query Arg \'{arg_name}\''



import copy
import logging
import re
from typing import (TYPE_CHECKING, Any, MutableMapping, MutableSequence,
                    Optional, Tuple, Union, cast)

from bson import ObjectId

from nawah.classes import Query
from nawah.config import Config
from nawah.enums import AttrType

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import NawahQuerySpecialGroup

logger = logging.getLogger('nawah')


def _compile_query(
    *,
    collection_name: str,
    attrs: MutableMapping[str, 'Attr'],
    query: Query,
) -> Tuple[
    Optional[int],
    Optional[int],
    MutableMapping[str, int],
    Optional[MutableSequence['NawahQuerySpecialGroup']],
    MutableSequence[Any],
]:
    aggregate_prefix: MutableSequence[Any] = []
    aggregate_suffix: MutableSequence[Any] = []
    aggregate_query: MutableSequence[Any] = [{'$match': {'$and': []}}]
    aggregate_match = aggregate_query[0]['$match']['$and']
    skip: Optional[int] = None
    limit: Optional[int] = None
    sort: MutableMapping[str, int] = {'_id': -1}
    group: Optional[MutableSequence['NawahQuerySpecialGroup']] = None
    logger.debug('attempting to process query: %s', query)

    query = copy.deepcopy(query)

    # Convert all values to _id attrs to ObjectId
    # for i in range(len(query['_id'])):
    #     query['_id'][i] = ObjectId(query['_id'][i])

    # for oper in ['ne', 'gt', 'gte', 'lt', 'lte']:
    #     for i in range(len(query[f'_id:${oper}'])):
    #         query[f'_id:${oper}'][i] = ObjectId(query[f'_id:${oper}'][i][f'${oper}'])

    # for oper in ['in', 'nin']:
    #     for i in range(len(query[f'_id:${oper}'])):
    #         query[f'_id:${oper}'][i] = [
    #             ObjectId(_id) for _id in query[f'_id:${oper}'][i][f'${oper}']
    #         ]

    # [DOC] Update variables per Doc Mode
    if '__deleted' not in query or query['__deleted'] is False:
        aggregate_prefix.append({'$match': {'__deleted': {'$exists': False}}})
    else:
        # [DOC] This condition is expanded to allow __deleted = True, __deleted = None to have del query[__deleted] be applied to both conditions
        if query['__deleted'][0] is True:
            aggregate_prefix.append({'$match': {'__deleted': {'$exists': True}}})
        del query['__deleted'][0]

    if '__create_draft' not in query or query['__create_draft'] is False:
        aggregate_prefix.append({'$match': {'__create_draft': {'$exists': False}}})
    else:
        if query['__create_draft'][0] is True:
            aggregate_prefix.append({'$match': {'__create_draft': {'$exists': True}}})
        del query['__create_draft'][0]

    if ('__update_draft' not in query and '__update_draft:$ne' not in query) or query[
        '__update_draft'
    ] is False:
        query_update_draft = False
        aggregate_prefix.append({'$match': {'__update_draft': {'$exists': False}}})
    else:
        query_update_draft = True
        aggregate_prefix.append({'$match': {'__update_draft': {'$exists': True}}})
        if '__update_draft' in query and isinstance(
            query['__update_draft'][0], ObjectId
        ):
            aggregate_prefix.append(
                {'$match': {'__update_draft': query['__update_draft'][0]}}
            )
        elif '__update_draft:$ne' in query and query['__update_draft:$ne'][0] is False:
            aggregate_prefix.append({'$match': {'__update_draft': {'$ne': False}}})
        try:
            del query['__update_draft'][0]
        except:
            del query['__update_draft:$ne'][0]

    # [DOC] Update variables per Query Special Args
    if '$skip' in query:
        skip = query['$skip']
        del query['$skip']
    if '$limit' in query:
        limit = query['$limit']
        del query['$limit']
    if '$sort' in query:
        sort = query['$sort']
        del query['$sort']
    if '$group' in query:
        group = query['$group']
        del query['$group']
    if '$search' in query:
        aggregate_prefix.insert(0, {'$match': {'$text': {'$search': query['$search']}}})
        project_query: MutableMapping[str, Any] = {
            attr: '$' + attr for attr in attrs.keys()
        }
        project_query['_id'] = '$_id'
        project_query['__score'] = {'$meta': 'textScore'}
        aggregate_suffix.append({'$project': project_query})
        aggregate_suffix.append({'$match': {'__score': {'$gt': 0.5}}})
        del query['$search']
    if '$geo_near' in query:
        aggregate_prefix.insert(
            0,
            {
                '$geoNear': {
                    'near': {
                        'type': 'Point',
                        'coordinates': query['$geo_near']['val'],
                    },
                    'distanceField': query['$geo_near']['attr'] + '.__distance',
                    'maxDistance': query['$geo_near']['dist'],
                    'spherical': True,
                }
            },
        )
        del query['$geo_near']

    for step in query:
        _compile_query_step(
            aggregate_prefix=aggregate_prefix,
            aggregate_suffix=aggregate_suffix,
            aggregate_match=aggregate_match,
            collection_name=collection_name,
            attrs=attrs,
            step=step,
        )

    if '$attrs' in query and type(query['$attrs']) == list:
        group_query = {
            '_id': '$_id',
            **{
                attr: {'$first': f'${attr}'}
                for attr in query['$attrs']
                if attr in attrs.keys()
            },
        }
        # [DOC] We need to expose __update_draft value if it is queried as this refers to the original doc to be updated
        if query_update_draft:
            group_query['__update_draft'] = {'$first': '$__update_draft'}
        aggregate_suffix.append({'$group': group_query})
    else:
        group_query = {
            '_id': '$_id',
            **{attr: {'$first': f'${attr}'} for attr in attrs.keys()},
        }
        # [DOC] We need to expose __update_draft value if it is queried as this refers to the original doc to be updated
        if query_update_draft:
            group_query['__update_draft'] = {'$first': '$__update_draft'}
        aggregate_suffix.append({'$group': group_query})

    logger.debug(
        f'processed query, aggregate_prefix:{aggregate_prefix}, aggregate_suffix:{aggregate_suffix}, aggregate_match:{aggregate_match}'
    )
    if len(aggregate_match) == 1:
        aggregate_query = [{'$match': aggregate_match[0]}]
    elif len(aggregate_match) == 0:
        aggregate_query = []

    aggregate_query = aggregate_prefix + aggregate_query + aggregate_suffix
    return (skip, limit, sort, group, aggregate_query)


def _compile_query_step(
    *,
    aggregate_prefix: MutableSequence[Any],
    aggregate_suffix: MutableSequence[Any],
    aggregate_match: MutableSequence[Any],
    collection_name: str,
    attrs: MutableMapping[str, 'Attr'],
    step: Union[MutableMapping[str, Any], MutableSequence[Any]],
) -> None:
    if type(step) == dict and len(step.keys()):  # type: ignore
        step = cast(MutableMapping[str, Any], step)
        child_aggregate_query: MutableMapping[str, Any] = {'$and': []}
        for attr in step.keys():
            if attr.startswith('__or'):
                child_child_aggregate_query: MutableMapping[str, Any] = {'$or': []}
                _compile_query_step(
                    aggregate_prefix=aggregate_prefix,
                    aggregate_suffix=aggregate_suffix,
                    aggregate_match=child_child_aggregate_query['$or'],
                    collection_name=collection_name,
                    attrs=attrs,
                    step=step[attr],
                )
                if len(child_child_aggregate_query['$or']) == 1:
                    child_aggregate_query['$and'].append(
                        child_child_aggregate_query['$or'][0]
                    )
                elif len(child_child_aggregate_query['$or']) > 1:
                    child_aggregate_query['$and'].append(
                        child_child_aggregate_query['$or']
                    )
            else:
                # [DOC] Add extn query when required
                if (
                    attr.find('.') != -1
                    and attr.split('.')[0] in attrs.keys()
                    and attrs[attr.split('.')[0]].extn
                ):
                    # [TODO] Check if this works with EXTN as Attr Type TYPE
                    step_attr = attr.split('.')[1]
                    step_attrs: MutableMapping[str, 'Attr'] = Config.modules[
                        attrs[attr.split('.')[0]].extn.module
                    ].attrs

                    # [DOC] Don't attempt to extn attr that is already extended
                    lookup_query = False
                    for stage in aggregate_prefix:
                        if (
                            '$lookup' in stage.keys()
                            and stage['$lookup']['as'] == attr.split('.')[0]
                        ):
                            lookup_query = True
                            break
                    if not lookup_query:
                        extn_collection = Config.modules[
                            attrs[attr.split('.')[0]].extn.module
                        ].collection
                        aggregate_prefix.append(
                            {
                                '$addFields': {
                                    attr.split('.')[0]: {
                                        '$toObjectId': f'${attr.split(".")[0]}'
                                    }
                                }
                            }
                        )
                        aggregate_prefix.append(
                            {
                                '$lookup': {
                                    'from': extn_collection,
                                    'localField': attr.split('.')[0],
                                    'foreignField': '_id',
                                    'as': attr.split('.')[0],
                                }
                            }
                        )
                        aggregate_prefix.append({'$unwind': f'${attr.split(".")[0]}'})
                        group_query: MutableMapping[str, Any] = {
                            attr: {'$first': f'${attr}'} for attr in attrs.keys()
                        }
                        group_query[attr.split('.')[0]] = {
                            '$first': f'${attr.split(".")[0]}._id'
                        }
                        group_query['_id'] = '$_id'
                        aggregate_suffix.append({'$group': group_query})
                else:
                    step_attr = attr
                    step_attrs = attrs

                # [DOC] Convert strings and lists of strings to ObjectId when required
                if step_attr == '_id':
                    try:
                        if isinstance(step[attr], str):
                            step[attr] = ObjectId(step[attr])
                        elif isinstance(step[attr], list):
                            step[attr] = [
                                ObjectId(child_attr) for child_attr in step[attr]
                            ]
                        elif (
                            isinstance(step[attr], dict) and '$in' in step[attr].keys()
                        ):
                            step[attr] = {
                                '$in': [
                                    ObjectId(child_attr)
                                    for child_attr in step[attr]['$in']
                                ]
                            }
                    except:
                        logger.warning(
                            f'Failed to convert attr to id type: {step[attr]}'
                        )
                # [DOC] Check for query oper
                if type(step[attr]) == dict:
                    if '$regex' in step[attr].keys():
                        step[attr] = {
                            '$regex': re.compile(
                                step[attr]['$regex'], re.RegexFlag.IGNORECASE
                            )
                        }

                if type(step[attr]) == dict and '$match' in step[attr].keys():
                    child_aggregate_query['$and'].append(step[attr]['$match'])
                else:
                    child_aggregate_query['$and'].append({attr: step[attr]})
        if len(child_aggregate_query['$and']) == 1:
            aggregate_match.append(child_aggregate_query['$and'][0])
        elif len(child_aggregate_query['$and']) > 1:
            aggregate_match.append(child_aggregate_query)
    elif type(step) == list and len(step):
        step = cast(MutableSequence[Any], step)
        child_aggregate_query = {'$or': []}
        for child_step in step:
            _compile_query_step(
                aggregate_prefix=aggregate_prefix,
                aggregate_suffix=aggregate_suffix,
                aggregate_match=child_aggregate_query['$or'],
                collection_name=collection_name,
                attrs=attrs,
                step=child_step,
            )
        if len(child_aggregate_query['$or']) == 1:
            aggregate_match.append(child_aggregate_query['$or'][0])
        elif len(child_aggregate_query['$or']) > 1:
            aggregate_match.append(child_aggregate_query)



'''Provides 'create' Base Function callable'''

import asyncio
import datetime
import logging
import sys
import traceback
from functools import partial
from typing import TYPE_CHECKING, Any, MutableSequence, Optional, cast

import nawah.data as Data
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call, expand_val, validate_doc

from .exceptions import (DuplicateUniqueException, NoDocCreatedException,
                         UtilityModuleDataCallException)

if TYPE_CHECKING:
    from asyncio.futures import Future

    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger('nawah')


async def create(
    *,
    module_name: str,
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    doc: 'NawahDoc',
    raise_no_success: Optional[bool],
) -> 'Results':
    '''Creates doc for a module'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='create'
        )

    # Expand dot-notated keys onto dicts
    doc = expand_val(doc=doc)
    # Deleted all extra doc args
    doc = {
        attr: doc[attr]
        for attr in ['_id', *module.attrs]
        if attr in doc and doc[attr] is not None
    }
    # Append host_add, user_agent, create_time, if present in attrs
    if (
        'user' in module.attrs
        and 'host_add' not in doc
        and env['session']
        and Event.ATTRS_DOC not in skip_events
    ):
        doc['user'] = env['session']['user']['_id']
    if 'create_time' in module.attrs:
        doc['create_time'] = datetime.datetime.utcnow().isoformat()
    if 'host_add' in module.attrs and 'host_add' not in doc:
        doc['host_add'] = env['REMOTE_ADDR']
    if 'user_agent' in module.attrs and 'user_agent' not in doc:
        doc['user_agent'] = env['HTTP_USER_AGENT']
    if Event.ATTRS_DOC not in skip_events:
        # Check presence and validate all attrs in doc args
        validate_doc(
            mode='create',
            doc=doc,
            attrs=module.attrs,
        )
        # Check unique_attrs
        if module.unique_attrs:
            unique_attrs_query: MutableSequence[Any] = [[]]
            for attr in module.unique_attrs:
                if isinstance(attr, str):
                    attr = cast(str, attr)
                    unique_attrs_query[0].append({attr: doc[attr]})
                elif isinstance(attr, tuple):
                    unique_attrs_query[0].append(
                        {child_attr: doc[child_attr] for child_attr in attr}
                    )
                # [TODO] Implement use of single-item dict with LITERAL Attr Type for dynamic unique check based on doc value
            unique_attrs_query.append({'$limit': 1})
            unique_results = await call(
                'base/read',
                module_name=module_name,
                skip_events=[Event.PERM],
                env=env,
                query=unique_attrs_query,
            )
            if unique_results['args']['count']:
                raise DuplicateUniqueException(unique_attrs=module.unique_attrs)

    # Check for counters, to update doc
    if module.counters:
        counter_name: str
        counter_count: int
        counter_value: str
        counter_locks: dict[str, str] = {}
        counter_create: list[str] = []
        counter_update: list[str] = []

        for attr_name, counter in module.counters.items():
            # Counter.counter can be str or callable. Per type figure out lock name
            if isinstance(counter.counter, str):
                counter_name = f'__COUNTER:{counter.counter}'
            elif callable(counter.counter):
                counter_name = counter.counter(doc)

            # Obtain lock for counter_name to prevent duplicates
            counter_lock_results = await call(
                'base/obtain_lock',
                module_name='setting',
                skip_events=[Event.PERM],
                env=env,
                doc={'tags': [counter_name], 'attempts': 3},
            )
            # Add lock_id to counter_locks to delete later
            counter_locks[counter_name] = counter_lock_results['args']['docs'][0]['_id']

            # Read counter value
            counter_results = await call(
                'setting/read',
                skip_events=[Event.PERM],
                env=env,
                query=[{'type': 'global', 'var': counter_name}],
            )
            # At this point, it is possible, counter was never created. Handle that scenario
            if counter_results['args']['count'] == 0:
                # Append counter_name to counter_create so later we create it instead of updating
                # its value
                counter_create.append(counter_name)
                # And, set counter_count explicitly to 1
                counter_count = 1
            else:
                # Otherwise, add to counter_update, and add to current val
                counter_update.append(counter_name)
                counter_count = counter_results['args']['docs'][0]['val'] + 1

            # Pass counter_count to Counter.pattern_formatter to get counter_value to be used in doc
            counter_value = counter.pattern_formatter(counter_count)
            doc[attr_name] = counter_value

    # Execute Data driver create
    results = await Data.create(env=env, collection_name=module.collection, doc=doc)

    # After successfully creating doc, create, update counters and delete locks
    # We will do this asynchronously so we don't hold response of call
    if module.counters:
        for counter_name in counter_create:
            counter_call_task = asyncio.create_task(
                call(
                    'setting/create',
                    skip_events=[Event.PERM],
                    env=env,
                    doc={
                        'type': 'global',
                        'var': counter_name,
                        'val_type': {'type': 'INT', 'args': {}},
                        'val': 1,
                    },
                    args={'raise_no_success': True},
                )
            )
            counter_call_task.add_done_callback(
                partial(
                    _counter_call_callback,
                    env=env,
                    counter_name=counter_name,
                    counter_lock=counter_locks[counter_name],
                )
            )

        for counter_name in counter_update:
            counter_call_task = asyncio.create_task(
                call(
                    'setting/update',
                    # Skip Event ATTRS_QUERY to avoid additional requirements in query
                    skip_events=[Event.PERM, Event.ATTRS_QUERY],
                    env=env,
                    query=[{'var': counter_name}],
                    doc={
                        'val': {'$add': 1},
                    },
                    args={'raise_no_success': True},
                )
            )
            counter_call_task.add_done_callback(
                partial(
                    _counter_call_callback,
                    env=env,
                    counter_name=counter_name,
                    counter_lock=counter_locks[counter_name],
                )
            )

    # create soft action is to only return the new created doc _id.
    if Event.SOFT in skip_events:
        read_results = await call(
            'base/read',
            module_name=module_name,
            skip_events=[Event.PERM],
            env=env,
            query=[[{'_id': results['docs'][0]}]],
        )
        results = read_results['args']

    if raise_no_success is True and results['count'] == 0:
        raise NoDocCreatedException(module_name=module_name)

    return {'status': 200, 'msg': f'Created {results["count"]} docs', 'args': results}


def _counter_call_callback(
    counter_call_task: 'Future',
    env: 'NawahEnv',
    counter_name: str,
    counter_lock: str,
):
    try:
        # Assert we have good call
        counter_call_task.result()
        # Then, delete counter_lock
        delete_lock_call_task = asyncio.create_task(
            call(
                'base/delete_lock',
                module_name='setting',
                skip_events=[Event.PERM],
                env=env,
                query=[{'_id': counter_lock}],
            )
        )
        delete_lock_call_task.add_done_callback(
            partial(
                _delete_lock_call_callback,
                counter_name=counter_name,
                counter_lock=counter_lock,
            )
        )
    except Exception:  # pylint: disable=broad-except
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logger.error(
            'Call \'setting/[create|update]\' for counter \'%s\' has failed',
            counter_name,
        )
        logger.error(
            'Lock will remain enact to prevent the counter from being reused while it is '
            'not created/updated'
        )
        logger.debug('Error traceback:')
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            logger.error('- %s', line)


def _delete_lock_call_callback(
    delete_lock_call_task: 'Future',
    counter_name: str,
    counter_lock: str,
):
    try:
        # Assert we have good call
        delete_lock_call_task.result()
    except Exception:  # pylint: disable=broad-except
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logger.error(
            'Call \'base/delete_lock\' for counter \'%s\' with \'_id\' \'%s\' has failed. DELETE '
            'MANUALLY NOW',
            counter_name,
            counter_lock,
        )
        logger.debug('Error traceback:')
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            logger.error('- %s', line)



'''Provides 'Func', 'Perm' dataclasses'''

from dataclasses import dataclass
from typing import (TYPE_CHECKING, Any, Awaitable, Callable, MutableMapping,
                    MutableSequence, Optional, Union)

if TYPE_CHECKING:
    from nawah.types import Results

    from ._analytics import Analytics
    from ._attr import Attr
    from ._cache import Cache
    from ._module import Module


@dataclass(kw_only=True)
class Func:
    '''Func dataclass serves role of defining a Nawah Module Function, which
    identifies set of configuration to expose a callable as an endpoint'''

    # pylint: disable=too-many-instance-attributes

    permissions: MutableSequence['Perm']
    analytics: Optional['Analytics'] = None
    cache: Optional['Cache'] = None
    cache_channels_update: Optional[MutableSequence[str]] = None
    cache_channels_reset: Optional[MutableSequence[str]] = None
    call_args: Optional[MutableMapping[str, 'Attr']] = None
    query_attrs: Optional[
        Union[MutableMapping[str, 'Attr'], MutableSequence[MutableMapping[str, 'Attr']]]
    ] = None
    doc_attrs: Optional[
        Union[MutableMapping[str, 'Attr'], MutableSequence[MutableMapping[str, 'Attr']]]
    ] = None
    strict_doc: bool = False
    get_func: bool = False
    post_func: bool = False
    callable: Optional[Callable[..., Awaitable['Results']]] = None
    desc: Optional[str] = None
    name: Optional[str] = None
    module: Optional['Module'] = None

    def __post_init__(self):
        # Force setting func_name, module to None
        self.name = self.module = None
        # Convert query_attrs, doc_attrs to list if defined as dict
        if isinstance(self.query_attrs, dict):
            self.query_attrs = [self.query_attrs]
        if isinstance(self.doc_attrs, dict):
            self.doc_attrs = [self.doc_attrs]


@dataclass(kw_only=True)
class Perm:
    '''Perm dataclass serves role of defining a Permissions Set, where Nawah
    attempts to match value of 'privilege' with current 'User' privileges,
    and use values for 'query_mod', 'doc_mod' to modify respective values'''

    privilege: str
    query_mod: Optional[
        Union[MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]]
    ] = None
    doc_mod: Optional[MutableMapping[str, Any]] = None

    def __post_init__(self):
        pass



'''Provides exceptions needed for configuring Nawah'''


class ConfigException(Exception):
    '''Raised by 'config_module', 'config_app' Utility when failed to process or apply runtime
    config'''



'''Provides 'Attr' dataclass'''

import datetime
from dataclasses import dataclass
from typing import (TYPE_CHECKING, Any, Callable, MutableMapping,
                    MutableSequence, Optional, Union)

from nawah.enums import AttrType

if TYPE_CHECKING:
    from ._default import Default
    from ._extn import Extn


@dataclass(kw_only=True)
class Attr:
    '''Attr dataclass serves role of defining an Attr Type Set, which
    instructs Nawah of allowed data type for a specific value'''

    # pylint: disable=invalid-name, redefined-builtin, too-many-public-methods

    desc: Optional[str]
    type: AttrType
    args: MutableMapping[str, Any]
    default: Optional['Default'] = None
    extn: Optional['Extn'] = None
    valid: bool = False

    def __post_init__(self):
        pass

    @staticmethod
    def ANY(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'ANY' '''
        return Attr(desc=desc, type=AttrType.ANY, args={})

    @staticmethod
    def ATTR(desc: str = None, type: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'ATTR' '''
        return Attr(desc=desc, type=AttrType.ATTR, args={'type': type})

    @staticmethod
    def ID(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'ID' '''
        return Attr(desc=desc, type=AttrType.ID, args={})

    @staticmethod
    def STR(pattern: str = None, desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'STR' '''
        return Attr(desc=desc, type=AttrType.STR, args={'pattern': pattern})

    @staticmethod
    def INT(
        ranges: MutableSequence[MutableSequence[float]] = None, desc: str = None
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'INT' '''
        return Attr(desc=desc, type=AttrType.INT, args={'ranges': ranges})

    @staticmethod
    def FLOAT(
        ranges: MutableSequence[MutableSequence[float]] = None, desc: str = None
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'FLOAT' '''
        return Attr(desc=desc, type=AttrType.FLOAT, args={'ranges': ranges})

    @staticmethod
    def BOOL(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'BOOL' '''
        return Attr(desc=desc, type=AttrType.BOOL, args={})

    @staticmethod
    def LOCALE(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'LOCALE' '''
        return Attr(desc=desc, type=AttrType.LOCALE, args={})

    @staticmethod
    def LOCALES(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'LOCALES' '''
        return Attr(desc=desc, type=AttrType.LOCALES, args={})

    @staticmethod
    def EMAIL(
        allowed_domains: MutableSequence[str] = None,
        disallowed_domains: MutableSequence[str] = None,
        strict_matching: bool = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'EMAIL' '''
        return Attr(
            desc=desc,
            type=AttrType.EMAIL,
            args={
                'allowed_domains': allowed_domains,
                'disallowed_domains': disallowed_domains,
                'strict_matching': strict_matching,
            },
        )

    @staticmethod
    def PHONE(codes: MutableSequence[str] = None, desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'PHONE' '''
        return Attr(desc=desc, type=AttrType.PHONE, args={'codes': codes})

    @staticmethod
    def IP(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'IP' '''
        return Attr(desc=desc, type=AttrType.IP, args={})

    @staticmethod
    def URI_WEB(
        allowed_domains: MutableSequence[str] = None,
        disallowed_domains: MutableSequence[str] = None,
        strict_matching: bool = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'URI_WEB' '''
        return Attr(
            desc=desc,
            type=AttrType.URI_WEB,
            args={
                'allowed_domains': allowed_domains,
                'disallowed_domains': disallowed_domains,
                'strict_matching': strict_matching,
            },
        )

    @staticmethod
    def DATETIME(
        ranges: MutableSequence[MutableSequence[datetime.datetime]] = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'DATETIME' '''
        return Attr(desc=desc, type=AttrType.DATETIME, args={'ranges': ranges})

    @staticmethod
    def DATE(
        ranges: MutableSequence[MutableSequence[datetime.date]] = None, desc: str = None
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'DATE' '''
        return Attr(desc=desc, type=AttrType.DATE, args={'ranges': ranges})

    @staticmethod
    def TIME(
        ranges: MutableSequence[MutableSequence[str]] = None, desc: str = None
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'TIME' '''
        return Attr(desc=desc, type=AttrType.TIME, args={'ranges': ranges})

    @staticmethod
    def FILE(
        types: MutableSequence[str] = None,
        ratio_ranges: MutableSequence[MutableSequence[float]] = None,
        dims_ranges: MutableSequence[MutableSequence[float]] = None,
        size_ranges: MutableSequence[MutableSequence[float]] = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'FILE' '''
        return Attr(
            desc=desc,
            type=AttrType.FILE,
            args={
                'types': types,
                'ratio_ranges': ratio_ranges,
                'dims_ranges': dims_ranges,
                'size_ranges': size_ranges,
            },
        )

    @staticmethod
    def GEO_POINT(desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'GEO_POINT' '''
        return Attr(desc=desc, type=AttrType.GEO_POINT, args={})

    @staticmethod
    def LIST(
        list: MutableSequence['Attr'],
        len_range: MutableSequence[float] = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'LIST' '''
        return Attr(
            desc=desc, type=AttrType.LIST, args={'list': list, 'len_range': len_range}
        )

    @staticmethod
    def KV_DICT(
        key: 'Attr',
        val: 'Attr',
        req: MutableSequence[str] = None,
        len_range: MutableSequence[float] = None,
        desc: str = None,
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'KV_DICT' '''
        return Attr(
            desc=desc,
            type=AttrType.KV_DICT,
            args={'key': key, 'val': val, 'req': req, 'len_range': len_range},
        )

    @staticmethod
    def TYPED_DICT(dict: MutableMapping[str, 'Attr'], desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'TYPED_DICT' '''
        return Attr(desc=desc, type=AttrType.TYPED_DICT, args={'dict': dict})

    @staticmethod
    def LITERAL(
        literal: MutableSequence[Union[str, int, float, bool]], desc: str = None
    ) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'LITERAL' '''
        return Attr(desc=desc, type=AttrType.LITERAL, args={'literal': literal})

    @staticmethod
    def UNION(union: MutableSequence['Attr'], desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'UNION' '''
        return Attr(desc=desc, type=AttrType.UNION, args={'union': union})

    @staticmethod
    def TYPE(type: Union[Callable, str], desc: str = None) -> 'Attr':
        '''Shorthand method to define 'Attr' object of type 'TYPE' '''
        return Attr(desc=desc, type=AttrType.TYPE, args={'type': type})



'''Provides tooling to test Nawah Modules'''

from ._mock import mock_doc, mock_env, mock_results, mock_session, mock_user
from ._setup_test import setup_test

__all__ = [
    'mock_doc',
    'mock_env',
    'mock_results',
    'mock_session',
    'mock_user',
    'setup_test',
]



'''Provides 'delete' Data Function'''

import logging
from typing import TYPE_CHECKING, MutableSequence, Union

from bson import ObjectId

from nawah.config import Config
from nawah.enums import DeleteStrategy
from nawah.exceptions import UnknownDeleteStrategyException

if TYPE_CHECKING:
    from nawah.types import NawahEnv, ResultsArgs

logger = logging.getLogger('nawah')


async def delete(
    *,
    env: 'NawahEnv',
    collection_name: str,
    docs: MutableSequence[Union[str, ObjectId]],
    strategy: DeleteStrategy,
) -> 'ResultsArgs':
    '''Deletes 'doc' from 'collection' using connection 'conn' in 'env' '''

    # Check strategy to cherrypick update, delete calls and system_docs
    if strategy in [DeleteStrategy.SOFT_SKIP_SYS, DeleteStrategy.SOFT_SYS]:
        if strategy == DeleteStrategy.SOFT_SKIP_SYS:
            del_docs = [
                ObjectId(doc) for doc in docs if ObjectId(doc) not in Config.sys.docs
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    'Skipped soft delete for system docs due to \'DELETE_SOFT_SKIP_SYS\' strategy'
                )
        else:
            logger.warning('Detected \'DELETE_SOFT_SYS\' strategy for delete call')
            del_docs = [ObjectId(doc) for doc in docs]
        # Perform update call on matching docs
        collection = env['conn'][Config.data_name][collection_name]
        update_doc = {'$set': {'__deleted': True}}
        # If using Azure Mongo service update docs one by one
        results = await collection.update_many({'_id': {'$in': docs}}, update_doc)
        update_count = results.modified_count
        return {'count': update_count, 'docs': [{'_id': doc} for doc in docs]}

    if strategy in [DeleteStrategy.FORCE_SKIP_SYS, DeleteStrategy.FORCE_SYS]:
        if strategy == DeleteStrategy.FORCE_SKIP_SYS:
            del_docs = [
                ObjectId(doc) for doc in docs if ObjectId(doc) not in Config.sys.docs
            ]
            if len(del_docs) != len(docs):
                logger.warning(
                    'Skipped soft delete for system docs due to \'DELETE_FORCE_SKIP_SYS\' strategy'
                )
        else:
            logger.warning('Detected \'DELETE_FORCE_SYS\' strategy for delete call')
            del_docs = [ObjectId(doc) for doc in docs]
        # Perform delete query on matching docs
        collection = env['conn'][Config.data_name][collection_name]
        results = await collection.delete_many({'_id': {'$in': del_docs}})
        delete_count = results.deleted_count
        return {'count': delete_count, 'docs': [{'_id': doc} for doc in docs]}

    raise UnknownDeleteStrategyException(f'DeleteStrategy \'{strategy}\' is unknown')
